#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/rswitch.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>

#define	DRIVER_VERSION		"$Revision: 1.3 $"


#define RSWITCH_MAJOR	241

#define CPLD_BASE_PHYS	0x04000000
#define rCPLD_INC	(*(volatile unsigned long *)(cpld_base_virt + 0x18000))

#define RRIGHT_EVENT	(1 << 16)
#define RLEFT_EVENT	(2 << 16)
#define KNOB_EVENT	(4 << 16)
#define ALL_EVENTS	(7 << 16)


#define NEW_CPLD

#undef DEBUG

static ssize_t rswitch_read(struct file *, char *, size_t, loff_t *);
static ssize_t rswitch_write(struct file *, const char *, size_t, loff_t *);
static unsigned int rswitch_poll(struct file *, struct poll_table_struct *);
static int rswitch_ioctl(struct inode *, struct file *, unsigned int, unsigned long);


static struct file_operations rswitch_fops = {
	owner:		THIS_MODULE,
	read:		rswitch_read,
	poll:		rswitch_poll,
	ioctl:		rswitch_ioctl,
	write:		rswitch_write,
};

static DECLARE_WAIT_QUEUE_HEAD(rswitch_wq);
static unsigned long cpld_base_virt;

static int knob_status;
static int angle;
static int rswitch_event;


static void rswitch_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	int val = rCPLD_INC & ALL_EVENTS;

#ifdef DEBUG
	printk("! 0x%x\n", val );
#endif
	if (val & RRIGHT_EVENT)
	{
		angle++;
#ifdef DEBUG
		printk("Switch was turned right: %d\n", angle);
#endif
	}
	if (val & RLEFT_EVENT)
	{
		angle--;
#ifdef DEBUG
		printk("Switch was turned left: %d\n", angle);
#endif
	}
	if (val & KNOB_EVENT)
	{
#ifndef NEW_CPLD
		knob_status = knob_status = 0;
#else
		knob_status = 1;
#endif
#ifdef DEBUG
		printk("Knob was %s\n", knob_status ? "pressed" : "released");
#endif
	}

#ifdef NEW_CPLD
		/* This would happen if the knob was released.
		 */
	if (! val)
	{
		knob_status = 0;
		rswitch_event |= KNOB_EVENT;
		wake_up_interruptible(&rswitch_wq);
	}
#endif

		/* There may be spurious interrupts sometimes - who knows?
		 */
	if (val)
	{
		rswitch_event |= val;
		wake_up_interruptible(&rswitch_wq);
#ifndef NEW_CPLD
		rCPLD_INC = val;
#else
			/* No need to ack knob event with the new CPLD.
			 */
		rCPLD_INC = (val & ~(KNOB_EVENT));
#endif
	}
}


static int __init rswitch_init(void)
{
	int res = -1;

	if (!(cpld_base_virt = (unsigned long)ioremap(CPLD_BASE_PHYS, 0x30000)))
	{
		printk("%s: Failed to map %dMb of CPLD space at 0x%x\n",
		       __FUNCTION__, CPLD_BASE_PHYS >> 20, CPLD_BASE_PHYS);
		goto Done;
	}

		/* Configure EINT4 as rising edge triggered.
		 */
	rEXTINT &= ~(7 << 16);
	rEXTINT |= (4 << 16);

		/* Configure PE4 pin as EINT4.
		 */
	rPECON &= ~(3 << 8);
	rPECON |= (2 << 8);

		/* This stuff is needed to get interrupts on rotation
		 * events, not only on knob-press.
		 * (Copied from the LCD initialization routine.)
		 */
	if (rLCDCON1 == 0)
	{
		rPCCON = (rPCCON & 0xFFFFFF00)| 0x000000AA;
		rPDCON = (rPDCON & 0xFFFFFF03)| 0x000000A8;
#if 0
		rPDCON = (rPDCON & 0xFFFFFF00)| 0x000000AA;
#endif
		rLCDCON2 = 0x000DC000;
		rLCDCON3 = 0x0051000A;
		rLCDCON4 = 0x00000001;
		rLCDCON5 = 0x00000440;
		rLCDCON1 = 0x00000B75;
	}

		/* Clear all events.
		 */
	rCPLD_INC = ALL_EVENTS;

	if (request_irq(IRQ_EINT4, rswitch_interrupt, 0, "rswitch", NULL))
	{
		printk("Failed to install an interrupt handler "
		       "for rotary switch IRQ%d\n", IRQ_EINT4);
		goto Done;
	}

		/* Clear all events (again, just in case).
		 */
	rCPLD_INC = ALL_EVENTS;

	res = register_chrdev(RSWITCH_MAJOR, "rswitch", &rswitch_fops);

	if (res)
	{
		printk("Failed to register rswitch driver. Major: %d\n",
		       RSWITCH_MAJOR);
		free_irq(IRQ_EINT4, NULL);
		res = -EIO;
		goto Done;
	}

	printk (KERN_INFO "TRAB rotary switch driver " DRIVER_VERSION "\n");
Done:
	return res;
}

static ssize_t
rswitch_read(struct file *filp, char *buf, size_t size, loff_t *offp)
{
	if (size < sizeof(int))
	{
		return -ENOMEM;
	}

	if (put_user(angle, (int *)buf))
	{
		return -EFAULT;
	}

	rswitch_event &= ~(RLEFT_EVENT | RRIGHT_EVENT);

	return sizeof(int);
}

static unsigned int
rswitch_poll(struct file *filp, struct poll_table_struct *wait)
{
	unsigned int res = 0;

	poll_wait(filp, &rswitch_wq, wait);
	if (rswitch_event & KNOB_EVENT)
	{
		res |= POLLPRI;
	}
	if ((rswitch_event & RRIGHT_EVENT) || (rswitch_event & RLEFT_EVENT))
	{
		res |= POLLIN | POLLRDNORM;
	}

	return res;
}

static int
rswitch_ioctl(struct inode *inode, struct file *filp, unsigned int command, unsigned long arg)
{
	int err = 0;

	switch (command)
	{
		case RSWITCH_SET_ANGLE:
			if (get_user(angle, (int *)arg))
			{
				err = -EFAULT;
			}
			break;
		case RSWITCH_GET_KNOB_STATE:
			rswitch_event &= ~KNOB_EVENT;
			if (put_user(knob_status, (int *)arg))
			{
				err = -EFAULT;
			}
			break;
		default:
			err = -ENOTTY;
			break;
	}
	return err;
}

static ssize_t
rswitch_write(struct file *filp, const char *buf, size_t size, loff_t * offp)
{
	return -ENOSYS;
}

#ifdef MODULE

void __exit rswitch_exit(void)
{
	int err;

	if ((err = unregister_chrdev(RSWITCH_MAJOR, "rswitch")))
	{
		printk(KERN_ERR "RSWITCH: failed to unregister driver (%d)\n",
		       err);
	}
	free_irq(IRQ_EINT4, NULL);
}

module_exit(rswitch_exit);

MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("Rotary incremental sensor");

#endif

module_init(rswitch_init);


	/* End of file.
	 */

