/*
 *	Real Time Clock interface for Linux on S3C2400
 *
 *	Copyright (c) 2002 Gary Jennejohn (gj@denx.de)
 *
 *	Based on rtc.c by Paul Gortmaker
 *	Based on sa1100-rtc.c by Nils Faerber
 *	Date/time conversion routines taken from arch/arm/kernel/time.c
 *			by Linus Torvalds and Russel King
 *		and the GNU C Library
 *	( ... I love the GPL ... just take what you need! ;)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <asm/bitops.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <linux/rtc.h>

#define	DRIVER_VERSION		"$Revision: 1.5 $"

#define TIMER_FREQ		64

/* Those are the bits from a classic RTC we want to mimic */
#define RTC_IRQF		0x80	/* any of the following 3 is active */
#define RTC_PF			0x40
#define RTC_AF			0x20
#define RTC_UF			0x10

static unsigned long rtc_status;
static unsigned long rtc_irq_data;
static unsigned long rtc_freq = 64;

static struct fasync_struct *rtc_async_queue;
static DECLARE_WAIT_QUEUE_HEAD(rtc_wait);

extern spinlock_t rtc_lock;

static const unsigned char days_in_mo[] =
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

#define is_leap(year) \
	((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

static void rtc_decode_time (struct rtc_time *tm, int doalm);
static void rtc_encode_time(struct rtc_time tm, int doalm);

/* flags to prevent having both the UPDATE and PERIODIC interrupts */
/* active concurrently since we only have on source for a timer */
/* interrupt on the S3C2400 :( */
static char hz_intr = 0, per_intr = 0;

/*
 * Converts seconds since 1970-01-01 00:00:00 to Gregorian date.
 */

static void decodetime (unsigned long t, struct rtc_time *tval)
{
	long days, month, year, rem;

	days = t / 86400;
	rem = t % 86400;
	tval->tm_hour = rem / 3600;
	rem %= 3600;
	tval->tm_min = rem / 60;
	tval->tm_sec = rem % 60;
	tval->tm_wday = (4 + days) % 7;

#define LEAPS_THRU_END_OF(y) ((y)/4 - (y)/100 + (y)/400)

	year = 1970 + days / 365;
	days -= ((year - 1970) * 365
			+ LEAPS_THRU_END_OF (year - 1)
			- LEAPS_THRU_END_OF (1970 - 1));
	if (days < 0) {
		year -= 1;
		days += 365 + is_leap(year);
	}
	tval->tm_year = year - 1900;
	tval->tm_yday = days + 1;

	month = 0;
	if (days >= 31) {
		days -= 31;
		month++;
		if (days >= (28 + is_leap(year))) {
			days -= (28 + is_leap(year));
			month++;
			while (days >= days_in_mo[month]) {
				days -= days_in_mo[month];
				month++;
			}
		}
	}
	tval->tm_mon = month;
	tval->tm_mday = days + 1;
}

/* this routine services only ALARM interrupts */
static void rtc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	/* update irq data & counter */
	rtc_irq_data |= (RTC_AF|RTC_IRQF);
	rtc_irq_data += 0x100;

	/* wake up waiting process */
	wake_up_interruptible(&rtc_wait);
	kill_fasync (&rtc_async_queue, SIGIO, POLL_IN);
}

/* this routine services only PERIODIC interrupts */
static void timer1_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	/*
	 * If we match for the first time, the periodic interrupt flag won't
	 * be set.  If it is, then we did wrap around (very unlikely but
	 * still possible) and compute the amount of missed periods.
	 * The match reg is updated only when the data is actually retrieved
	 * to avoid unnecessary interrupts.
	 */
	if (hz_intr)
		rtc_irq_data |= (0x100|RTC_UF|RTC_IRQF);
	else
		rtc_irq_data |= (0x100|RTC_PF|RTC_IRQF);

	wake_up_interruptible(&rtc_wait);
	kill_fasync (&rtc_async_queue, SIGIO, POLL_IN);
}

static int rtc_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit (1, &rtc_status))
		return -EBUSY;
	rtc_irq_data = 0;
	return 0;
}

static int rtc_release(struct inode *inode, struct file *file)
{
	spin_lock_irq (&rtc_lock);
	rRTCCON = 1;
	rTICNT = 0;
	rRTCALM = 0;
	rRTCCON = 0;
	/* in case there are pending interrupts */
	rINTPND = (BIT_TICK|BIT_RTC);
	rSRCPND = (BIT_TICK|BIT_RTC);
	spin_unlock_irq (&rtc_lock);
	rtc_status = 0;
	return 0;
}

static int rtc_fasync (int fd, struct file *filp, int on)
{
	return fasync_helper (fd, filp, on, &rtc_async_queue);
}

static unsigned int rtc_poll(struct file *file, poll_table *wait)
{
	poll_wait (file, &rtc_wait, wait);
	return (rtc_irq_data) ? 0 : POLLIN | POLLRDNORM;
}

static loff_t rtc_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

ssize_t rtc_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long data;
	ssize_t retval;

	if (count < sizeof(unsigned long))
		return -EINVAL;

	add_wait_queue(&rtc_wait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	for (;;) {
		spin_lock_irq (&rtc_lock);
		data = rtc_irq_data;
		if (data != 0) {
			rtc_irq_data = 0;
			break;
		}
		spin_unlock_irq (&rtc_lock);

		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			goto out;
		}

		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto out;
		}

		schedule();
	}

	spin_unlock_irq (&rtc_lock);

	data -= 0x100;	/* the first IRQ wasn't actually missed */

	retval = put_user(data, (unsigned long *)buf);
	if (!retval)
		retval = sizeof(unsigned long);

out:
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&rtc_wait, &wait);
	return retval;
}

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	struct rtc_time tm, tm2;

	switch (cmd) {
	/* turn off ALARM interrupts */
	case RTC_AIE_OFF:
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		rRTCALM = 0;
		rRTCCON = 0;
		rtc_irq_data = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* turn on ALARM interrupts */
	case RTC_AIE_ON:
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		/* turn on all the bits */
		rRTCALM = 0x7f;
		rRTCCON = 0;
		rtc_irq_data = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* turn off UPDATE (1 Hz) interrupts */
	case RTC_UIE_OFF:
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		rTICNT &= ~(1 << 7);
		rRTCCON = 0;
		rtc_irq_data = 0;
		hz_intr = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* turn on UPDATE (1 Hz) interrupts */
	/* not supported */
	case RTC_UIE_ON:
	/* although this conflicts with periodic interrupts it seems */
	/* to be required ! */
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		/* 1 HZ */
		rTICNT = 127;
		rTICNT |= (1 << 7);
		rRTCCON = 0;
		rtc_irq_data = 0;
		hz_intr = 1;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* turn off PERIODIC interrupts */
	case RTC_PIE_OFF:
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		rTICNT &= ~(1 << 7);
		rRTCCON = 0;
		rtc_irq_data = 0;
		per_intr = 1;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* turn on PERIODIC interrupts */
	/* this conflicts with update interrupts */
	case RTC_PIE_ON:
		if (hz_intr != 0)
			return -EBUSY;
		if ((rtc_freq > 64) && !capable(CAP_SYS_RESOURCE))
			return -EACCES;
		spin_lock_irq(&rtc_lock);
		rRTCCON = 1;
		/* periodic interrupts are at (n+1)/128, n = 1 to 127 */
		/* so this is only accurate to 1/128 of a second */
		rTICNT = 128/rtc_freq - 1;
		rTICNT |= (1 << 7);
		rRTCCON = 0;
		rtc_irq_data = 0;
		per_intr = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	/* read the current ALARM values */
	case RTC_ALM_READ:
		rtc_decode_time (&tm, 1);
		break;
	/* set the current ALARM values */
	case RTC_ALM_SET:
		if (copy_from_user (&tm2, (struct rtc_time*)arg, sizeof (tm2)))
			return -EFAULT;
		rtc_encode_time(tm2, 1);
		return 0;
	/* read the current time */
	case RTC_RD_TIME:
		rtc_decode_time (&tm, 0);
		break;
	/* set the current time */
	case RTC_SET_TIME:
		if (!capable(CAP_SYS_TIME))
			return -EACCES;
		if (copy_from_user (&tm, (struct rtc_time*)arg, sizeof (tm)))
			return -EFAULT;
		if (tm.tm_year < 0 || (unsigned)tm.tm_mon >= 12 ||
		    tm.tm_mday < 1 || tm.tm_mday > (days_in_mo[tm.tm_mon] +
				(tm.tm_mon == 1 && is_leap(tm.tm_year))) ||
		    (unsigned)tm.tm_hour >= 24 ||
		    (unsigned)tm.tm_min >= 60 ||
		    (unsigned)tm.tm_sec >= 60)
			return -EINVAL;
		rtc_encode_time(tm, 0);
		return 0;
	/* read the current PERIODIC interrupt value */
	case RTC_IRQP_READ:
		return put_user(rtc_freq, (unsigned long *)arg);
	/* set the current PERIODIC interrupt value */
	/* note that on the S3C2400 the maximum frequency is only 64 HZ ! */
	case RTC_IRQP_SET:
		if (arg < 1 || arg > TIMER_FREQ)
			        return -EINVAL;
		if ((arg > 64) && (!capable(CAP_SYS_RESOURCE)))
			        return -EACCES;
		rtc_freq = arg;
		return 0;
	case RTC_EPOCH_READ:
		return put_user (1970, (unsigned long *)arg);
	default:
		return -EINVAL;
	}
	return copy_to_user ((void *)arg, &tm, sizeof (tm)) ? -EFAULT : 0;
}

static struct file_operations rtc_fops = {
	owner:		THIS_MODULE,
	llseek:		rtc_llseek,
	read:		rtc_read,
	poll:		rtc_poll,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release,
	fasync:		rtc_fasync,
};

static struct miscdevice s3c2400rtc_miscdev = {
	RTC_MINOR,
	"rtc",
	&rtc_fops
};

static int rtc_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	char *p = page;
	int len;
	struct rtc_time tm;

	rtc_decode_time (&tm, 0);
	p += sprintf(p, "rtc_time\t: %02d:%02d:%02d\n"
			"rtc_date\t: %04d-%02d-%02d\n"
			"rtc_epoch\t: %04d\n",
			tm.tm_hour, tm.tm_min, tm.tm_sec,
			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, 1970);
	rtc_decode_time (&tm, 1);
	p += sprintf(p, "alrm_time\t: %02d:%02d:%02d\n"
			"alrm_date\t: %04d-%02d-%02d\n",
			tm.tm_hour, tm.tm_min, tm.tm_sec,
			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);
	p += sprintf(p, "alarm_IRQ\t: %s\n", (rRTCALM & 0x7f) ? "yes" : "no" );
	p += sprintf(p, "periodic_IRQ\t: %s\n", (per_intr == 1) ? "yes" : "no");
	p += sprintf(p, "update_IRQ\t: %s\n", (hz_intr == 1) ? "yes" : "no");
	p += sprintf(p, "periodic_freq\t: %ld\n", rtc_freq);

	len = (p - page) - off;
	if (len < 0)
		len = 0;

	*eof = (len <= count) ? 1 : 0;
	*start = page + off;

	return len;
}

#ifndef FROM_BCD
#define FROM_BCD(bcd, dec) (dec) = ((((bcd & 0xf0) >> 4) * 10) + ((bcd) & 0x0f))
#define TO_BCD(bcd, dec) (bcd) = ((((dec)/10) << 4) + ((dec) % 10))
#endif
/* these 2 routines decode/encode the BCD timer registers to/from */
/* a decimal value. */
static void rtc_decode_time (struct rtc_time *tm, int doalm)
{
	unsigned char scr;
	unsigned long flags;

	spin_lock_irqsave (&rtc_lock, flags);
	/* enable accesses */
	rRTCCON |= 1;
	if (doalm)
		scr = rALMSEC;
	else
		scr = rBCDSEC;
	FROM_BCD(scr, tm->tm_sec);
	if (doalm)
		scr = rALMMIN;
	else
		scr = rBCDMIN;
	FROM_BCD(scr, tm->tm_min);
	if (doalm)
		scr = rALMHOUR;
	else
		scr = rBCDHOUR;
	FROM_BCD(scr, tm->tm_hour);
	/* this is the day of the month */
	if (doalm)
		scr = rALMDAY;
	else
		scr = rBCDDAY;
	FROM_BCD(scr, tm->tm_mday);
	/* this is the weekday */
	/* ALARM does not use the weekday */
	if (!doalm)
	{
		scr = rBCDDATE;
		FROM_BCD(scr, tm->tm_wday);
	}
	if (doalm)
		scr = rALMMON;
	else
		scr = rBCDMON;
	FROM_BCD(scr, tm->tm_mon);
	if (tm->tm_mon)
		tm->tm_mon -= 1;
	if (doalm)
		scr = rALMYEAR;
	else
		scr = rBCDYEAR;
	FROM_BCD(scr, tm->tm_year);
	/* hack to convert to >= 2000 since dates are relative to 1900 */
	if (tm->tm_year < 70)
		tm->tm_year += 100;
	/* disable accesses */
	rRTCCON &= ~1;
	spin_unlock_irqrestore (&rtc_lock, flags);
}

static void rtc_encode_time(struct rtc_time tm, int doalm)
{
	unsigned char scr;

	spin_lock_irq (&rtc_lock);
	/* enable accesses */
	rRTCCON |= 1;
	TO_BCD(scr, tm.tm_sec);
	if (doalm)
		rALMSEC = scr;
	else
		rBCDSEC = scr;
	TO_BCD(scr, tm.tm_min);
	if (doalm)
		rALMMIN = scr;
	else
		rBCDMIN = scr;
	TO_BCD(scr, tm.tm_hour);
	if (doalm)
		rALMHOUR = scr;
	else
		rBCDHOUR = scr;
	/* this is the day of the month */
	TO_BCD(scr, tm.tm_mday);
	if (doalm)
		rALMDAY = scr;
	else
		rBCDDAY = scr;
	/* this is the weekday - not used by ALARM */
	TO_BCD(scr, tm.tm_wday);
	if (!doalm)
		rBCDDATE = scr;
	tm.tm_mon += 1;
	TO_BCD(scr, tm.tm_mon);
	if (doalm)
		rALMMON = scr;
	else
		rBCDMON = scr;
	/* hack to convert to >= 2000 since dates are relative to 1900 */
	if (tm.tm_year >= 100)
		tm.tm_year -= 100;
	TO_BCD(scr, tm.tm_year);
	if (doalm)
		rALMYEAR = scr;
	else
		rBCDYEAR = scr;
	/* disable accesses */
	rRTCCON &= ~1;
	spin_unlock_irq (&rtc_lock);
}


/* this is all a real PITA because system time is kept as seconds */
/* since the epoc and the RTC uses only BCD encoded values. */
unsigned long s3c2400_get_rtc_time(void)
{
	struct rtc_time tm;

	rtc_decode_time(&tm, 0);
	return mktime (tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
				tm.tm_hour, tm.tm_min, tm.tm_sec);
}

int s3c2400_set_rtc(void)
{
	unsigned long current_time = xtime.tv_sec;
	struct rtc_time tm;

	if (rRTCALM & 0x7f) {
		unsigned long alarm_time, rtc_time;

		rtc_decode_time(&tm, 1);
		tm.tm_mon += 1;
		alarm_time = mktime (tm.tm_year + 1900, tm.tm_mon + 1,
			tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
		rtc_time = s3c2400_get_rtc_time();
		/* make sure not to forward the clock over an alarm */
		if (current_time >= alarm_time && alarm_time >= rtc_time)
			return -ERESTARTSYS;
	}
	decodetime(current_time, &tm);
	rtc_encode_time(tm, 0);
	return 0;
}

static int __init rtc_init(void)
{
	int ret;

	misc_register (&s3c2400rtc_miscdev);
	create_proc_read_entry ("driver/rtc", 0, 0, rtc_read_proc, NULL);
	/* unmask RTC interrupts */
	rINTMSK &= ~(BIT_TICK|BIT_RTC);

	ret = request_irq (IRQ_RTC, rtc_interrupt, SA_INTERRUPT, "rtc Alrm", NULL);
	if (ret) {
		printk(KERN_ERR "rtc: IRQ %d already in use.\n", IRQ_RTC);
		goto IRQ_RTCAlrm_failed;
	}
	ret = request_irq (IRQ_TICK, timer1_interrupt, SA_INTERRUPT, "rtc timer", NULL);
	if (ret) {
		printk(KERN_ERR "rtc: IRQ %d already in use.\n", IRQ_TICK);
		goto IRQ_OST1_failed;
	}

	printk (KERN_INFO "S3C2400 Real Time Clock driver " DRIVER_VERSION "\n");

	return 0;

IRQ_OST1_failed:
	free_irq (IRQ_RTC, NULL);
IRQ_RTCAlrm_failed:
	remove_proc_entry ("driver/rtc", NULL);
	misc_deregister (&s3c2400rtc_miscdev);
	return ret;
}

static void __exit rtc_exit(void)
{
	/* mask RTC interrupts */
	rINTMSK |= (BIT_TICK|BIT_RTC);
	free_irq (IRQ_TICK, NULL);
	free_irq (IRQ_RTC, NULL);
	remove_proc_entry ("driver/rtc", NULL);
	misc_deregister (&s3c2400rtc_miscdev);
}

module_init(rtc_init);
module_exit(rtc_exit);

MODULE_AUTHOR("Gary Jennejohn (gj@denx.de)");
MODULE_DESCRIPTION("S3C2400 Realtime Clock Driver (RTC)");
EXPORT_NO_SYMBOLS;
