/*
 *  linux/drivers/char/trab_rs485.c
 *
 *  Driver for S3C2400 serial port use fro RS485
 *
 *  Copyright (C) 2002 Gary Jennejohn (gj@denx.de)
 *
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *  Based on drivers/serial/serial_amba.c
 *  Copyright 1999 ARM Limited
 *  Copyright (C) 2000 Deep Blue Solutions Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: trab_rs485.c,v 1.10 2002/11/15 20:35:56 wd Exp $
 *
 * This is a driver for RS485 support on the TRAB board. It uses
 * UART1 of the internal S3C2400 UARTs.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/skbuff.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/version.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/trab_rs485.h>
#include <linux/trab_callout.h>

#define TRAB_RS485_MAJOR	240

#define	DRIVER_VERSION		"$Revision: 1.10 $"

/* declarations for fops */
static int trab_rs485_open(struct inode *, struct file *);
static int trab_rs485_close(struct inode *, struct file *);
static unsigned int trab_rs485_poll(struct file *, struct poll_table_struct *);
static int trab_rs485_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static ssize_t trab_rs485_read (struct file *, char *, size_t , loff_t *);
static ssize_t trab_rs485_write (struct file *, const char *, size_t , loff_t *);
static void trab_rs485_start_xmit(void);
static void trab_rs485_tx_chars(void);
static void trab_rs485_handle_timer(void *);
static void trab_rs485_do_xmit(void *);
extern void udelay(long delay);

/* -------------------
   Fops-struct 
   ------------------- */
static struct file_operations trab_rs485_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	owner: THIS_MODULE,
#endif
	read: trab_rs485_read,
	write: trab_rs485_write,
	poll: trab_rs485_poll,  /* poll */
	ioctl: trab_rs485_ioctl, /* ioctl */
	open: trab_rs485_open,  /* open */
	release: trab_rs485_close, /* release */
};

static unsigned int trab_rs485_fclk = 0;
static unsigned int trab_rs485_uartclk = 0;
/* queues to hold wdog accesses - this has priority */
static struct sk_buff_head wdogq;
/* queues to hold data read or to be written */
static struct sk_buff_head readq, writeq;
/* max length of the wrtie Q */
#define MAX_WQ_LEN 10
/* wait Qs for reading and polling */
static wait_queue_head_t rwait, pwait;
/* state of transfer - NONE, XMIT or RCV */
/* this is needed because the MAX3845 is half-duplex */
#define STATE_NONE	1
/* In Progress */
#define STATE_XMIT_IP	2
#define STATE_RCV_IP	4
/* Wait Completion */
#define STATE_XMIT_WC	8
#define STATE_RCV_WC	16
/* receive is the next state */
#define STATE_RETRY	32
static atomic_t xfer_state;

/* RS485 is half-duplex, so we only need one structure for xmit/rcv */
static struct rs485_packet rcv_packet, xmit_packet;
static struct rs485_packet *active_packet;
static struct rs485_data *active_data;
static struct sk_buff *active_skb;
/* stuff associated with the watchdog activities */
static int wdog_timer_active = 0;
static struct rs485_wdog *wdog_head;
/* a watchdog transfer is active */
static int doing_wdog;
/* number of nodes with a changed status */
static int wdog_statchanged;
static unsigned long rx_start; /* start of rcv in jiffies */
static int last_error;
/* callout list handle */
static struct callout_list *callout_tx, *callout_tr;
/* period in which the application must re-trigger the watchdogs, */
/* in seconds. */
static int wdog_period_int;
static unsigned long wdog_period_jiffies;

/*
 * Access macros for the S3C2400 UARTs
 */
#define membase (SERBASE + 0x4000)
#define GET_ULCON()	*((unsigned int *)(membase + ULCON))
#define GET_UCON()	*((unsigned int *)(membase + UCON))
#define GET_UFCON()	*((unsigned int *)(membase + UFCON))
#define GET_UMCON()	*((unsigned int *)(membase + UMCON))
#define GET_UTRSTAT()	*((unsigned int *)(membase + UTRSTAT))
#define GET_UERSTAT()	*((unsigned int *)(membase + UERSTAT))
#define GET_UFSTAT()	*((unsigned int *)(membase + UFSTAT))
#define GET_UMSTAT()	*((unsigned int *)(membase + UMSTAT))
#define GET_URXH()	*((unsigned int *)(membase + URXH))
#define GET_UBRDIV()	*((unsigned int *)(membase + UBRDIV))
#define SET_ULCON(val)	*((unsigned int *)(membase + ULCON)) = (val)
#define SET_UCON(val) *((unsigned int *)(membase + UCON)) = (val)
#define SET_UFCON(val) *((unsigned int *)(membase + UFCON)) = (val)
#define SET_UMCON(val) *((unsigned int *)(membase + UMCON)) = (val)
#define SET_UTRSTAT(val) *((unsigned int *)(membase + UTRSTAT)) = (val)
#define SET_UERSTAT(val) *((unsigned int *)(membase + UERSTAT)) = (val)
#define SET_UFSTAT(val) *((unsigned int *)(membase + UFSTAT)) = (val)
#define SET_UMSTAT(val) *((unsigned int *)(membase + UMSTAT)) = (val)
#define SET_UTXH(val) *((unsigned int *)(membase + UTXH)) = (val)
#define SET_UBRDIV(val) *((unsigned int *)(membase + UBRDIV)) = (val)

/* interrupt mask manipulations */
#define ENABLE_IRQ(n) enable_irq((n))
#define DISABLE_IRQ(n) disable_irq((n))

#undef HANG_DEBUG
#undef TIMER_DEBUG
#undef RETRY_DEBUG
#if defined(HANG_DEBUG) || defined(TIMER_DEBUG) || defined(RETRY_DEBUG)
/* for debugging */
static char rs485_serdebug[512];
static char *rs485_sbptr = rs485_serdebug;
#endif
#ifdef HANG_DEBUG
#define SERDEBUG(f) \
#ifdef TIMER_DEBUG
#define TIMERERDEBUG(f) \
#endif
rs485_sbptr += sprintf(rs485_sbptr, f); \
if ((rs485_sbptr - rs485_serdebug) > 510) \
rs485_sbptr = rs485_serdebug;
#else
#define SERDEBUG(f)
#endif
#define cpld_rs485_re   (*(volatile ulong *)io_p2v(0x04028000))
/* define this to use the FIFOs - does not seem to work correctly */
#undef USE_FIFOS

/* used to count opens since Linux does it wrong */
static atomic_t rs485_open_count;

/* atomic_set() and atomic_read() are not atomic !!! */
#define my_atomic_read(x, v) { \
	unsigned long flags; \
	save_flags(flags); \
	cli(); \
	(x) = ((v)->counter); \
	restore_flags(flags); \
}

#define my_atomic_set(v, i) { \
	unsigned long flags; \
	save_flags(flags); \
	cli(); \
	((v)->counter) = (i); \
	restore_flags(flags); \
}


/*
 * Calculate, intelligently, the CRC of a dataset incrementally given a
 * buffer full at a time.
 * Initialize crc to 0 for XMODEM, -1 for CCITT.
 *
 * Usage:
 *   newcrc = updcrc( oldcrc, bufadr, buflen )
 *        unsigned int oldcrc, buflen;
 *        char *bufadr;
 *
 */

    /* the CRC polynomial. This is used by XMODEM (almost CCITT).
     */
    /* number of bits in CRC: don't change it. */
#define W 16

    /* this the number of bits per char: don't change it. */
#define B 8

static unsigned short crctab[1<<B] = { /* as calculated by initcrctab() */
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
 };


/*
 * The idea here is that data are put into readq as they come in
 * and then consumed as necessary by trab_rs485_read().
 */
static ssize_t
trab_rs485_read (struct file *filp, char *buffer, size_t size, loff_t *off)
{
	ssize_t N;
	struct sk_buff *skb;

	DECLARE_WAITQUEUE(wait, current);

	if (size == 0)
		return (0);

	__add_wait_queue(&rwait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	do {
		if (!skb_queue_empty (&readq))
			break;

		if (filp->f_flags & O_NONBLOCK)
		{
			N = -EAGAIN;
			goto out;
		}
                if (signal_pending(current)) {
			N = -ERESTARTSYS;
			goto out;
		}
		schedule();
	} while (1);

	skb = skb_dequeue (&readq);
	if (skb->len < size)
		N = skb->len;
	else
		N = size;

	if (copy_to_user (buffer, skb->data, N))
		N = -EFAULT;

	kfree_skb (skb);

  out:
	current->state = TASK_RUNNING;
	__remove_wait_queue(&rwait, &wait);

	return (N);
}

static ssize_t
trab_rs485_write (struct file *filp, const char *buffer, size_t size, loff_t *off)
{
	unsigned char *data;
	struct sk_buff *skb;

	if (skb_queue_len(&writeq) > MAX_WQ_LEN)
		return (0);

	if (size != sizeof(struct rs485_data))
	{
		printk("%s: invalid size %d\n", __FUNCTION__, size);
		return -EINVAL;
	}

	if ((skb = alloc_skb (size, GFP_KERNEL)) == NULL)
		return (-ENOMEM);

	data = skb_put (skb, size);

	/* copy data in from user */
	if (copy_from_user (data, buffer, size))
	{
		kfree_skb(skb);
		return (-EFAULT);
	}
	/* put it in the write Q */
	skb_queue_tail (&writeq, skb);

	/* start the transmission */
	trab_rs485_start_xmit();

	return(size);
}

static int
trab_rs485_open(struct inode *inode, struct file *file)
{
	/* XXX */
	static char trab_rs485_timer_inited = 0;

	file->f_op = &trab_rs485_fops;

	/* do this only once at first open */
	if (!trab_rs485_timer_inited)
	{
		my_atomic_set(&rs485_open_count, 0);
		trab_rs485_timer_inited = 1;
		wdog_timer_active = 1;
		callout_tr = add_callout(trab_rs485_handle_timer, NULL, 1, 1);
	}
	atomic_inc(&rs485_open_count);
	MOD_INC_USE_COUNT;
	return(0);
}

static int
trab_rs485_close(struct inode *inode, struct file *file)
{

	if (atomic_dec_and_test(&rs485_open_count))
	{
		unsigned long flags;
		struct rs485_wdog *wp, *nwp;
		/* last close - flush queues and stop the wdog */
		__save_flags_cli(flags);
		for (wp = wdog_head; wp != NULL; wp = nwp) {
			nwp = wp->next;
			kfree_skb(wp);
		}
		wdog_head = NULL;
		doing_wdog = 0;
		wdog_statchanged = 0;
		wdog_period_int = 0;
		wdog_period_jiffies = 0;
		skb_queue_purge (&readq);
		skb_queue_purge (&writeq);
		skb_queue_purge (&wdogq);
		__restore_flags(flags);
	}
	MOD_DEC_USE_COUNT;
	return(0);
}

static unsigned int
trab_rs485_poll(struct file *file, struct poll_table_struct *pota)
{
	unsigned int mask = 0;
	struct rs485_wdog *wp;

	/* The write queue is limited to 10 entries */
	if (skb_queue_len(&writeq) < MAX_WQ_LEN)
		mask |= POLLOUT | POLLWRNORM;

	poll_wait (file, &pwait, pota);

	/* check for readable */
	if (!skb_queue_empty (&readq))
		mask |= POLLIN | POLLRDNORM;

	/* check for a change in status after a watchdog poll. */
	/* this cannot report which node's status changed. */
	/* this shows up in the exception mask. */
	for (wp = wdog_head; wp != NULL; wp = wp->next) {
		if (wp->statch != 0) {
			mask |= POLLPRI;
			break;
		}
	}

	return (mask);
}

static int
trab_rs485_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long data)
{
	struct rs485_wdog *wp, *lnk, **dlnk;
	struct rs485_data *rd;
	unsigned long flags;

	switch(cmd) {
	case RS485_WDOG_PERIOD:
		/* period is in seconds !! */
		if (copy_from_user((void *)&wdog_period_int, (void *)data,
				sizeof(int)))
			return -EFAULT;
		/* assumes that HZ is 100 !! */
		__save_flags_cli(flags);
		wdog_period_jiffies = jiffies + 100 * wdog_period_int;
		__restore_flags(flags);
		break;
	case RS485_WDOG_TRIGGER:
		/* assumes that HZ is 100 !! */
		__save_flags_cli(flags);
		wdog_period_jiffies = jiffies + 100 * wdog_period_int;
		__restore_flags(flags);
		break;
	case RS485_SET_WDOG:
		/* allocate space for the data from the user */
		wp = (struct rs485_wdog *)kmalloc(sizeof(*wp), GFP_KERNEL);
		if (wp == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)wp, (void *)data, sizeof(*wp)))
			return -EFAULT;
		/* timo is in milliseconds and each jiffie is 10 ms */
		wp->jiffies = jiffies + wp->timo/10;
		wp->statch = 0;
		wp->next = NULL;
		/* add it to the linked list - do not allow duplicates */
		for (dlnk = &wdog_head; *dlnk != NULL; dlnk = &(*dlnk)->next) {
			if ((*dlnk)->rs485_wdog_data.target_node ==
				wp->rs485_wdog_data.target_node) {
				wp->next = (*dlnk)->next;
				*dlnk = wp;
				break;
			}
		}
		if (*dlnk == NULL)
			*dlnk = wp;
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "dlnk-%p wh-%p ", dlnk, wdog_head);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
		break;
	case RS485_GET_COUNT:
		/* calculate this every time */
		wdog_statchanged = 0;
		for (lnk = wdog_head; lnk != NULL; lnk = lnk->next) {
			if (lnk->statch != 0)
				wdog_statchanged++;
		}
		if (copy_to_user((void *)data, (void *)&wdog_statchanged,
			sizeof(int)))
			return -EFAULT;
		break;
	case RS485_GET_STATUS:
		/* NB: the user passes in a (struct rs485_data *) !! */
		/* we can only do one at a time */
		for (lnk = wdog_head; lnk != NULL; lnk = lnk->next) {
			if (lnk->statch != 0) {
				if (copy_to_user((void *)data,
					(void *)&lnk->rs485_wdog_data,
					sizeof(*rd)))
					return -EFAULT;
				lnk->statch = 0;
				wdog_statchanged--;
				break;
			}
		}
		break;
#ifdef GJDEBUG
	case RS485_RESET_STATE:
		trab_rs485_disable_tx();
		active_data = active_packet = NULL;
		break;
#endif
	default:
		return -EINVAL;
	}
	return(0);
}

#define MPLLCON (io_p2v(0x14800004))
#define CLKDIVN (io_p2v(0x14800014))
static void __init
trab_rs485_get_fclk(void)
{
	unsigned int mpllcon = *(unsigned int *)MPLLCON;
	int clkdivn = (*(unsigned int *)CLKDIVN) & 3;
	int clk_factor[4] = {1, 2, 2, 4};

	/* the S3C2400 uses a 12 MHz xtal */
	/* formula from page 6-22 */
	trab_rs485_fclk = ((12000000 * (8 + ((mpllcon & 0xff000) >> 12)))/((2 + ((mpllcon & 0x1f0) >> 4)) * (1 << (mpllcon & 3))));
	trab_rs485_uartclk = trab_rs485_fclk/clk_factor[clkdivn];
}

static unsigned short
updcrc(unsigned char *icp, unsigned int icnt )
{
	/* this is CCITT */
	unsigned short crc = 0xffff;
	unsigned char *cp = icp;
	unsigned int cnt = icnt;
	while( cnt-- )
		crc = (crc<<B) ^ crctab[(crc>>(W-B)) ^ *cp++];

	return( crc );
}

/*
 * State table:
 *	RE	DE	Result
 *	1	1	XMIT
 *	0	0	RCV
 *	1	0	Shutdown
 */

/* function that controls the receiver enable for the rs485 */
/* rs485re_state reflects the level (0/1) of the RE pin */
static void
set_rs485re(unsigned char rs485re_state)
{
	if(rs485re_state)
		cpld_rs485_re = 0x010000;
	else
		cpld_rs485_re = 0x0;
}

/* function that controls the sender enable for the rs485 */
/* rs485de_state reflects the level (0/1) of the DE pin */
static void
set_rs485de(unsigned char rs485de_state)
{
	unsigned long padat = rPADAT;

	/* This is on PORT A bit 11 */
	if(rs485de_state)
		padat |= (1 << 11);
	else
		padat &= ~(1 << 11);
	rPADAT = padat;
}

/*
 * Update the watchdog timeout for the node. If we do a normal access
 * to the node then the watchdog is automatically satisfied.
 */
static void
wdog_update(void)
{
	struct rs485_wdog *wp;
	unsigned long flags;

	/* don't want a timer interrupt here */
	save_flags(flags);
	cli();
	for (wp = wdog_head; wp != NULL; wp = wp->next) {
		if (wp->rs485_wdog_data.target_node == active_data->target_node)
			/* timo is in milliseconds and each jiffie is 10 ms */
			wp->jiffies = jiffies + wp->timo/10;
	}
	restore_flags(flags);
}

/*
 * Check whether the status changed since the last watchdog access to
 * the node.  If the status data changed, copy the new data from active_data,
 * set the changed flag (statch) and wake up any process which may be
 * waiting in poll/select.
 */
static void
wdog_check_statchange(void)
{
	struct rs485_wdog *wp;
	unsigned long flags;

	/* don't want a timer interrupt here */
	save_flags(flags);
	cli();
	doing_wdog = 0;
	for (wp = wdog_head; wp != NULL; wp = wp->next) {
		if (wp->rs485_wdog_data.target_node == active_data->target_node)
		{
			/* memcmp returns !0 if the data differ */
			if (memcmp(wp->rs485_wdog_data.rx_data,
			    active_data->rx_data,
			    wp->rs485_wdog_data.rx_expected) != 0) {
				/* to from length */
				memcpy((void *)wp->rs485_wdog_data.rx_data,
					(void *)active_data->rx_data,
					wp->rs485_wdog_data.rx_expected);
				wp->statch = 1;
				wake_up_interruptible(&pwait);
			}
		}
	}
	restore_flags(flags);
}

static void
trab_rs485_disable_tx(void)
{
	set_rs485de(0);
	my_atomic_set(&xfer_state, STATE_NONE);
	DISABLE_IRQ(IRQ_UTXD1);
}

static void
trab_rs485_enable_tx(void)
{
	set_rs485de(1);
	set_rs485re(1);
	my_atomic_set(&xfer_state, STATE_XMIT_IP);
	ENABLE_IRQ(IRQ_UTXD1);
}

static void
trab_rs485_disable_rx(void)
{
	set_rs485re(1);
	my_atomic_set(&xfer_state, STATE_NONE);
	DISABLE_IRQ(IRQ_URXD1);
}

static void
trab_rs485_enable_rx(void)
{
	set_rs485re(0);
	set_rs485de(0);
	my_atomic_set(&xfer_state, STATE_RCV_IP);
	ENABLE_IRQ(IRQ_URXD1);
}

static int
trab_rs485_tx_empty(void)
{
	int status;

#ifdef USE_FIFOS
	status = GET_UFSTAT() & (UFSTAT_TXCNTMASK | UFSTAT_TXFIFOFULL);
	return status ? 0 : 1;
#else
	status = GET_UTRSTAT() & (UTRSTAT_TXBSMPTY);
	return status ? 1 : 0;
#endif
}

#ifdef USE_OVERRUN
static void
trab_rs485_check_rx_overrun(void)
{
	unsigned int ch, rsr;

	rsr = GET_UERSTAT();
	if ((rsr & UERSTAT_OVERRUN))
	{
		ch = GET_URXH();
	}
}
#endif

static void
trab_rs485_rx_chars(void)
{
	unsigned char *data;
	unsigned int ch = 0, rsr = 0;
	static unsigned int global_rsr = 0;
	unsigned short crc;
	struct sk_buff *skb;
	unsigned long flags;

	/* don't want an interrupt here */
	save_flags(flags);
	cli();
	/*
	 * This is a pain. How can I tell whether the wrong number
	 * (too few, too many) of bytes is received ?!?
	 */
	data = (unsigned char *)active_packet;
#ifdef USE_FIFOS
	while ((GET_UFSTAT() & (UFSTAT_RXCNTMASK | UFSTAT_RXFIFOFULL)) > 0) {
#else
	while (GET_UTRSTAT() & UTRSTAT_RXDATA) {
#endif
		ch = GET_URXH();
		/* we received a character so reset the timer */
		/* protect this with a spinlock ? */
		rx_start = jiffies;

SERDEBUG("R");
#ifdef TIMER_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "R ");
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
#ifdef HANG_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "%#x ", ch);
if ((rs485_sbptr - rs485_serdebug) > 510)
rs485_sbptr = rs485_serdebug;
#endif
		rsr = GET_UERSTAT();

		/*
		 * ANY error is fatal
		 */
		if (rsr & UERSTAT_ANY) {
			global_rsr = rsr & UERSTAT_ANY;
			rSRCPND = BIT_UERR01;
			rINTPND = BIT_UERR01;
			break;
		} else
			rsr = 0;
		data[active_packet->offset++] = ch;
		/* this will catch too many bytes, but too few ? */
		--active_packet->count;
	}
	restore_flags(flags);
	/* not finished yet */
	if (active_packet->count > 0)
		return;

	/* error - FATAL */
	if (global_rsr != 0) {
#ifdef DEBUG_OUTPUT
		printk("%s serial error %#x - retry\n", __FUNCTION__, global_rsr);
#endif
		last_error = RESULT_SER;
		goto errout;
	}
	/* don't want an interrupt here */
	save_flags(flags);
	cli();
	rx_start = 0;
	restore_flags(flags);

#undef RXDEBUG
#if defined(RXDEBUG) && defined(HANG_DEBUG)
	{
		int i;
		for (i= 0; i < active_packet->offset; i++) {
			rs485_sbptr += sprintf(rs485_sbptr, "%#x ", data[i]);
			if ((rs485_sbptr - rs485_serdebug) > 510)
				rs485_sbptr = rs485_serdebug;
		}
	}
#endif
#undef RXDEBUG
	trab_rs485_disable_rx();
	/* may want to turn this into a kernel thread */
	if (RCV_OK(active_data, active_packet)) {
		/* CRC is for the preamble, command and the data */
		crc = updcrc(data, active_packet->offset - 2);
		if (CRC_OK(crc, active_data, active_packet)) {
			/* to from length */
			/* THIS WAS BACKWARDS */
			memcpy(active_data->rx_data, active_packet->data,
				active_data->rx_expected);
			last_error = active_data->result = RESULT_OK;
			if (doing_wdog) {
				wdog_check_statchange();
			} else {
				skb = dev_alloc_skb(sizeof(struct rs485_data));
				if (skb == NULL) {
					printk("%s - no memory for sk_buff!\n", __FUNCTION__);
					/* XXX - this is wrong */
					goto errout;
				}
				data = skb_put(skb, sizeof(struct rs485_data));
				/* to from length */
				memcpy(data, active_data, sizeof(struct rs485_data));
				skb_queue_tail(&readq, skb);
#ifdef TIMER_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "W1 ");
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "lj-%ld ", jiffies);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
				/* wakeup the process waiting on read() */
				/* and select */
				wake_up_interruptible(&rwait);
				wake_up_interruptible(&pwait);
			}
			/* start a new xmit */
			trab_rs485_start_xmit();
		} else {
#ifdef DEBUG_OUTPUT
			printk("%s CRC error - retry\n", __FUNCTION__);
#endif
			last_error = RESULT_CRC;
			goto errout;
		}
	} else {
#ifdef DEBUG_OUTPUT
		printk("%s data error - retry\n", __FUNCTION__);
#endif
		last_error = RESULT_SER;
		goto errout;
	}
	global_rsr = 0;
	return;
errout:
	global_rsr = 0;
	trab_rs485_disable_rx();
	my_atomic_set(&xfer_state, STATE_RETRY);
	trab_rs485_start_xmit();
	return;
}

static int retry_cnt = 0;
static void
trab_rs485_start_xmit()
{
	/* the protocol states that min. 1.3 ms must pass between */
	/* transmissions. We actually wait between 10 and 20 ms */
	callout_tx = add_callout(trab_rs485_do_xmit, NULL, 2, 0);
	/* couldn't add the callout, just call do_xmit() */
	if (callout_tx == NULL)
		trab_rs485_do_xmit(NULL);
}

static void
trab_rs485_do_xmit(void *junk)
{
	unsigned short crc;
	unsigned long flags;
	int res;

	save_flags(flags);
	cli();

	/* already protected, don't need the macro */
	res = ((&xfer_state)->counter);
	if (res != STATE_NONE)
		if (res != STATE_RETRY)
			return;

	/* if this is a retry then use the old active_data */
	/* otherwise start fresh. */
	if (res != STATE_RETRY) {
		retry_cnt = 0;
		/* avoid memory leaks */
		if (active_skb != NULL)
			kfree_skb(active_skb);
		active_packet = NULL;
		active_data = NULL;
		active_skb = NULL;
		if (skb_queue_empty(&writeq) && skb_queue_empty(&wdogq))
		{
			restore_flags(flags);
			return;
		}
		/* the watchdog has priority */
		if (!skb_queue_empty(&wdogq)) {
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "WDS ");
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
			active_skb = skb_dequeue (&wdogq);
			doing_wdog = 1;
		} else {
			if (skb_queue_empty(&writeq))
			{
				restore_flags(flags);
				return;
			}
			active_skb = skb_dequeue (&writeq);
			doing_wdog = 0;
		}
		active_data = (struct rs485_data *)active_skb->data;
	}

	/* the ';' is already part of the MACROs */
	SET_MS(&xmit_packet)
	SET_COUNT(active_data, &xmit_packet)
	SET_COMMAND(active_data, &xmit_packet)
	SET_NODE(active_data, &xmit_packet)
	/* +4 for preamble, command and CRC */
	xmit_packet.count = active_data->tx_count + 4;
	xmit_packet.offset = 0;
	/* to from length */
	memcpy(xmit_packet.data, active_data->tx_data, xmit_packet.count);
	/* CRC is for the preamble, command and the data */
	crc = updcrc((unsigned char *)&xmit_packet, xmit_packet.count - 2);
	SET_CRC(crc, active_data, &xmit_packet)
	wdog_update();
	/* start the transmission */
	active_packet = &xmit_packet;
	trab_rs485_enable_tx();
	restore_flags(flags);
	trab_rs485_tx_chars();
}

/* define this to get transmit debugging */
#undef TXDEBUG
static void
trab_rs485_tx_chars(void)
{
	unsigned char *data;
	unsigned long flags;
	int res;
#ifdef TXDEBUG
	static int data_dumped = 0;
#endif

	/* This requires that the last byte transmitted causes an interrupt */
	if (active_packet == NULL) {
		my_atomic_read(res, &xfer_state);
		if (res == STATE_XMIT_WC) {
			/* wait for the xmit to complete */
			while (trab_rs485_tx_empty() == 0) {
				udelay(100);
				SERDEBUG("W");
			}
SERDEBUG("Y");
			active_data->result = RESULT_OK;
			/* don't use the old data */
			memset((unsigned char *)&rcv_packet, 0,
				sizeof(rcv_packet));
			memset(active_data->rx_data, 0, MAX_BYTES);
			active_packet = &rcv_packet;
			/* +4 for preamble, command and CRC */
			active_packet->count = active_data->rx_expected + 4;
			active_packet->offset = 0;
			trab_rs485_disable_tx();
			trab_rs485_enable_rx();
			/* don't want an interrupt here */
			save_flags(flags);
			cli();
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "fj-%ld ", jiffies);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
			rx_start = jiffies;
			restore_flags(flags);
		}
		return;
	}

	data = (unsigned char *)active_packet;
	if ( data == NULL || data != (unsigned char *)&xmit_packet)
		return;
#if defined(TXDEBUG) && defined(HANG_DEBUG)
	if (data_dumped == 0)
	{
		int i;
		for (i= 0; i < active_packet->count; i++) {
			rs485_sbptr += sprintf(rs485_sbptr, "%#x ", data[i]);
			if ((rs485_sbptr - rs485_serdebug) > 510)
				rs485_sbptr = rs485_serdebug;
		}
		data_dumped = 1;
	}
#endif
#ifdef USE_FIFOS
	while ((GET_UFSTAT() & UFSTAT_TXFIFOFULL) != UFSTAT_TXFIFOFULL) {
#else
	while ((GET_UTRSTAT() & UTRSTAT_TXBMPTY)) {
#endif
		SET_UTXH(data[active_packet->offset++]);
		if (--active_packet->count == 0)
			break;
	}
	if (active_packet->count == 0) {
SERDEBUG("Z");
		my_atomic_set(&xfer_state, STATE_XMIT_WC);
		active_packet = NULL;
#ifdef TXDEBUG
		data_dumped = 0;
#endif
	}
}

static void
trab_rs485_tx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	trab_rs485_tx_chars();
}

static void
trab_rs485_rx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	trab_rs485_rx_chars();
}

static void
trab_rs485_change_speed(u_int baudrate)
{
	u_int quot = 0;

	/* Set baud rate */
	quot = (trab_rs485_uartclk/(16 * baudrate)) - 1;
	SET_UBRDIV(quot);
}

static void
trab_rs485_handle_timer(void *not_used)
{
	struct sk_buff *skb;
	unsigned char *data;
	struct rs485_wdog *wp;
	unsigned long flags, last_rx;
	int res;

	/* don't want an interrupt here */
	save_flags(flags);
	cli();
	/* the watchdog's watchdog timeout */
	if (wdog_period_jiffies < jiffies)
		wdog_period_jiffies = 0;
	last_rx = rx_start;
	restore_flags(flags);
	/* check the receive timeout timer */
	if (last_rx != 0) {
#ifdef TIMER_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "j-%ld r-%ld ", jiffies, last_rx);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
		/* not timed out yet */
		if ((jiffies - last_rx) < 2)
			return;
		rx_start = 0;
		/* safety first */
		my_atomic_read(res, &xfer_state);
		if (res != STATE_RCV_IP || active_packet == NULL
			|| active_data == NULL)
			return;
#ifdef TIMER_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "c-%d ", retry_cnt);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
		/* limit the number of retries to 5 */
		if (retry_cnt++ > 3) {
#ifdef DEBUG_OUTPUT
			printk("%s too many retries\n", __FUNCTION__);
#endif
			/* this also sets xfer_state */
			trab_rs485_disable_rx();
			retry_cnt = 0;
			active_data->result = last_error;
			if (last_error == RESULT_OK)
				active_data->result = RESULT_TO;
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "lj-%ld ", jiffies);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
			if (doing_wdog) {
				wdog_check_statchange();
			} else {
				skb = dev_alloc_skb(sizeof(struct rs485_data));
				if (skb == NULL) {
					printk("%s - no memory for retry sk_buff!\n",
						__FUNCTION__);
					panic("no retry sk_buff");
				}
				data = skb_put(skb, sizeof(struct rs485_data));
				/* to from length */
				memcpy(data, active_data, sizeof(struct rs485_data));
				skb_queue_tail(&readq, skb);
				/* wakeup the process waiting on read() */
				wake_up_interruptible(&rwait);
				wake_up_interruptible(&pwait);
			}
		} else {
			/* do a retry */
			trab_rs485_disable_rx();
			my_atomic_set(&xfer_state, STATE_RETRY);
			trab_rs485_start_xmit();
			return;
		}
	}

	/* no watchdogs if the application hasn't triggered us */
	if (wdog_period_jiffies == 0)
		goto out;
	/* this is where we do the watchdog stuff */
	for (wp = wdog_head; wp != NULL; wp = wp->next) {
		if (wp->jiffies <= jiffies) {
			skb = dev_alloc_skb(sizeof(struct rs485_data));
			if (skb == NULL) {
				printk("%s - no memory for wdog sk_buff!\n",
					__FUNCTION__);
				panic("no wdog sk_buff");
			}
			data = skb_put(skb, sizeof(struct rs485_data));
			/* to from length */
			memcpy(data, &wp->rs485_wdog_data,
				sizeof(struct rs485_data));
			skb_queue_tail(&wdogq, skb);
#ifdef RETRY_DEBUG
rs485_sbptr += sprintf(rs485_sbptr, "WDA wj-%ld j-%ld", wp->jiffies, jiffies);
if ((rs485_sbptr - rs485_serdebug) > 508)
rs485_sbptr = rs485_serdebug;
#endif
		}
	}
    out:
	/* does nothing if a transfer is already in progress */
	trab_rs485_start_xmit();
}

static int __init
trab_rs485_init(void)
{
	int retval;
	u_int ufcon, ucon;
	unsigned long pacon;

	if (trab_rs485_fclk == 0 || trab_rs485_uartclk == 0)
		trab_rs485_get_fclk();

	/* register rs485 device */
	if ((retval = register_chrdev(TRAB_RS485_MAJOR, "rs485", &trab_rs485_fops)))
	{
		printk("*** %s: unable to get major %d for RS485 dev\n", __FUNCTION__, TRAB_RS485_MAJOR);
		return retval;
	}

	/*
	 * Allocate the IRQs
	 */
	retval = request_irq(IRQ_URXD1, trab_rs485_rx_int, 0, "rs485_rx", NULL);
	if (retval)
	{
		unregister_chrdev(TRAB_RS485_MAJOR, "rs485");
		return retval;
	}
	retval = request_irq(IRQ_UTXD1, trab_rs485_tx_int, 0, "rs485_tx", NULL);
	if (retval)
	{
		unregister_chrdev(TRAB_RS485_MAJOR, "rs485");
		free_irq(IRQ_URXD1, NULL);
		return retval;
	}

	/*
	 * Finally, enable interrupts
	 */
	/* no interrupts on errors */
	rINTMSK |= BIT_UERR01;
	/* block the IRQ */
    	DISABLE_IRQ(IRQ_URXD1);
    	DISABLE_IRQ(IRQ_UTXD1);

	/* default baud rate is 38400 */
	trab_rs485_change_speed(38400);

#ifdef USE_FIFOS
	/* enable the FIFOs */
	ufcon = (UFCON_FIFO|UFCON_RXFIFOTL12|UFCON_TXFIFOTL4);
	ufcon |= (UFCON_RXFIFORESET|UFCON_TXFIFORESET);
#else
	ufcon = 0;
#endif
	/* RS485 requires a pulsed interrupt, NOT level ! */
	ucon = (UCON_RXINT|UCON_TXINT);
#ifdef USE_FIFOS
	ucon |= UCON_RXTO;
#endif
	SET_UCON(ucon);
	SET_UFCON(ufcon);
#ifdef USE_FIFOS
	/* wait for the FIFOs to reset */
	while ((GET_UFCON() & (UFCON_RXFIFORESET|UFCON_TXFIFORESET)) != 0) ;
#endif

	/* set 8N1 */
	SET_ULCON(ULCON_8BITS);

	/* RXD1|TXD1 */
	rPFCON |= ((2<<2)|(2<<6));
	rPFUP |= ((1<<1)|(1<<3));

	/* make PA11 an output - RS485_DE */
	pacon = rPACON;
	pacon &= ~(1 << 11);
	rPACON = pacon;
	/* set RS485_RE to 1 */
	set_rs485re(1);

	/* set xmit and rcv to known states */
	/* NOTE: the MAX3845 will be in power-down state after this ! */
	trab_rs485_disable_tx();
	trab_rs485_disable_rx();

	skb_queue_head_init(&wdogq);
	skb_queue_head_init(&readq);
	skb_queue_head_init(&writeq);
	init_waitqueue_head(&rwait);
	init_waitqueue_head(&pwait);
	active_packet = NULL;
	active_data = NULL;
	active_skb = NULL;
	wdog_head = NULL;
	callout_tr = NULL;
	callout_tx = NULL;
	doing_wdog = 0;
	rx_start = 0;
	wdog_statchanged = 0;
	wdog_period_int = 0;
	wdog_period_jiffies = 0;
	last_error = RESULT_OK;

	rSRCPND = (BIT_URXD1|BIT_UTXD1);
	rINTPND = (BIT_URXD1|BIT_UTXD1);
	/* free the IRQ */
    	ENABLE_IRQ(IRQ_URXD1);
    	ENABLE_IRQ(IRQ_UTXD1);
	printk("TRAB RS485 driver " DRIVER_VERSION " initialized\n");
	return 0;
}

module_init(trab_rs485_init);

#ifdef MODULE
/* **************************************************************************
  
    Function:    cleanup_module
  
    Description: Unregister the RS485-Driver
    
    return:      ---
                 
  
   *************************************************************************** */
void
cleanup_module(void)
{
	struct rs485_wdog *wp, *nwp;

	skb_queue_purge (&readq);
	skb_queue_purge (&writeq);
	skb_queue_purge (&wdogq);
	if (wdog_timer_active)
		rmv_callout(callout_tr);
	unregister_chrdev(TRAB_RS485_MAJOR, "rs485");
	free_irq(IRQ_URXD1, NULL);
	free_irq(IRQ_UTXD1, NULL);
	for (wp = wdog_head; wp != NULL; wp = nwp) {
		nwp = wp->next;
		kfree_skb(wp);
	}
	if (callout_tr != NULL)
		rmv_callout(callout_tr);
	if (callout_tx != NULL)
		rmv_callout(callout_tx);
}
#endif /* MODULE */

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Gary Jennejohn");
MODULE_DESCRIPTION("TRAB RS485 driver");
MODULE_LICENSE("GPL");
