/*
 * SRAM driver for TRAB board
 * Copyright (C) 2002 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>

#include <asm/uaccess.h>

#define TRAB_SRAM_MAJOR	32

 /* arch/arm/mm/mm-s3c2400.c
  */
#define SRAM_PHYS	0x02000000
#define SRAM_START	0xd0800000
#define SRAM_SIZE	0x00040000
#define SRAM_END	(SRAM_START + SRAM_SIZE)


static ssize_t sram_read(struct file * file, char * buf,
			size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	int i;
	
	if (p >= SRAM_SIZE)
		return 0;

	if (count > SRAM_SIZE - p)
		count = SRAM_SIZE - p;

	p += SRAM_START;

	/*
	 * We can't use simple copy_to_user here due to the
	 * S3C2400X bug with loading exactly 2 registers from
	 * non-cached memory using the LDM instruction (depending
	 * on its arguments, the memcpy routine may do exactly
	 * that).
	 */
#if 0
	if (copy_to_user(buf, (void*) p, count))
		return -EFAULT;
#else
	for (i = 0; i < count; i ++)
		if (put_user(*(char*)(p++), buf++))
			return -EFAULT;
#endif

	*ppos += count;

	return count;
}

static ssize_t sram_write(struct file * file, const char * buf, 
			 size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	int i;

	if (p >= SRAM_SIZE)
		return 0;

	if (count > SRAM_SIZE - p)
		count = SRAM_SIZE - p;

	p += SRAM_START;

	/*
	 * We can't use simple copy_from_user here due to the
	 * S3C2400X bug with loading exactly 2 registers from
	 * non-cached memory using the LDM instruction (depending
	 * on its arguments, the memcpy routine may do exactly
	 * that).
	 */
#if 0
	if (copy_from_user((void*) p, buf, count))
		return -EFAULT;
#else
	for (i = 0; i < count; i ++)
		if (get_user(*(char*)(p++), buf++))
			return -EFAULT;
#endif

	*ppos += count;

	return count;
}

static int sram_mmap(struct file * file, struct vm_area_struct * vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long size = vma->vm_end - vma->vm_start;

	if (offset + size > SRAM_SIZE)
		return -EINVAL;

	offset += SRAM_PHYS;

	if (remap_page_range(vma->vm_start, offset, size, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static loff_t sram_lseek(struct file * file, loff_t offset, int orig)
{
	switch (orig) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		case 2:
			file->f_pos = SRAM_SIZE - offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static int sram_open(struct inode * inode, struct file * filp)
{
	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

static struct file_operations sram_fops = {
	llseek:		sram_lseek,
	read:		sram_read,
	write:		sram_write,
	mmap:		sram_mmap,
	open:		sram_open,
};

int __init trab_sram_init(void)
{
	if (register_chrdev (TRAB_SRAM_MAJOR, "sram", & sram_fops)) 
	{
		printk (KERN_ERR "trab_sram: register_chrdev failed\n");
		return -EIO;
	}

	return 0;
}

/******************************
 **** Module Declarations *****
 **************************** */

module_init (trab_sram_init);

#ifdef MODULE

static void trab_sram_cleanup (void)
{
	if (unregister_chrdev (TRAB_SRAM_MAJOR, "sram")) 
	{
		printk (KERN_ERR "trab_sram: failed to unregister character device\n");
	}
}

module_exit (trab_sram_cleanup);

#endif	/* MODULE */
