/*
 * Driver for A/D converters, thermo and pressure sensors on TRAB board.
 * Copyright (C) 2002 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/stat.h>
#include <linux/wait.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>

#include <linux/tsc2000.h>
#include "Pt1000_voltage_data.h"
#include "Pt1000_temp_data.h"
#include "TypK_temp_data.h"

/* #define DEBUG */

#define ADC_DEBUG

/* define this to use the new CPLD software with separate PENIRQ/DAV */
#define NEWCPLD

#include <linux/trab_callout.h>

#define	DRIVER_VERSION		"$Revision: 1.14 $"

#define SET_CS_TOUCH 		(rPDDAT &= 0x5FF)
#define CLR_CS_TOUCH 		(rPDDAT |= 0x200)

#ifdef DEBUG
#define DEBUG_SET_PDDAT_0	(rPDDAT |= 0x1)
#define DEBUG_CLR_PDDAT_0	(rPDDAT &= ~0x1)
#define DEBUG_SET_PDDAT_4	(rPDDAT |= 0x10)
#define DEBUG_CLR_PDDAT_4	(rPDDAT &= ~0x10)
#else
#define DEBUG_SET_PDDAT_0
#define DEBUG_CLR_PDDAT_0
#define DEBUG_SET_PDDAT_4
#define DEBUG_CLR_PDDAT_4
#endif

#define CON_MUX0         	(rPCCON = (rPCCON & 0x0FFFFFCFF) | 0x00000100)
#define CLR_MUX0         	(rPCDAT &= 0x0FFEF)
#define SET_MUX0         	(rPCDAT |= 0x00010)
 
#define CON_MUX1         	(rPCCON = (rPCCON & 0x0FFFFF3FF) | 0x00000400)
#define CLR_MUX1         	(rPCDAT &= 0x0FFDF)
#define SET_MUX1         	(rPCDAT |= 0x00020)
 
#define CON_MUX1_ENABLE  	(rPCCON = (rPCCON & 0x0FFFFCFFF) | 0x00001000)
#define CLR_MUX1_ENABLE  	(rPCDAT |= 0x00040)
#define SET_MUX1_ENABLE  	(rPCDAT &= 0x0FFBF)
 
#define CON_MUX2_ENABLE  	(rPCCON = (rPCCON & 0x0FFFF3FFF) | 0x00004000)
#define CLR_MUX2_ENABLE  	(rPCDAT |= 0x00080)
#define SET_MUX2_ENABLE  	(rPCDAT &= 0x0FF7F)
 
#define CON_MUX3_ENABLE  	(rPCCON = (rPCCON & 0x0FFFCFFFF) | 0x00010000)
#define CLR_MUX3_ENABLE  	(rPCDAT |= 0x00100)
#define SET_MUX3_ENABLE  	(rPCDAT &= 0x0FEFF)
 
#define CON_MUX4_ENABLE  	(rPCCON = (rPCCON & 0x0FFF3FFFF) | 0x00040000)
#define CLR_MUX4_ENABLE  	(rPCDAT |= 0x00200)
#define SET_MUX4_ENABLE  	(rPCDAT &= 0x0FDFF)

#define CON_SEL_TEMP_V_0 	(rPCCON = (rPCCON & 0x0FFCFFFFF) | 0x00100000)
#define CLR_SEL_TEMP_V_0 	(rPCDAT &= 0x0FBFF)
#define SET_SEL_TEMP_V_0 	(rPCDAT |= 0x00400)

#define CON_SEL_TEMP_V_1 	(rPCCON = (rPCCON & 0x0FF3FFFFF) | 0x00400000)
#define CLR_SEL_TEMP_V_1 	(rPCDAT &= 0x0F7FF)
#define SET_SEL_TEMP_V_1 	(rPCDAT |= 0x00800)

#define CON_SEL_TEMP_V_2 	(rPCCON = (rPCCON & 0x0FCFFFFFF) | 0x01000000)
#define CLR_SEL_TEMP_V_2 	(rPCDAT &= 0x0EFFF)
#define SET_SEL_TEMP_V_2 	(rPCDAT |= 0x01000)

#define CON_SEL_TEMP_V_3 	(rPCCON = (rPCCON & 0x0F3FFFFFF) | 0x04000000)
#define CLR_SEL_TEMP_V_3 	(rPCDAT &= 0x0DFFF)
#define SET_SEL_TEMP_V_3 	(rPCDAT |= 0x02000)

#define TSC2000_REG_X		((0 << 11) | (0 << 5))
#define TSC2000_REG_Y		((0 << 11) | (1 << 5))
#define TSC2000_REG_Z1		((0 << 11) | (2 << 5))
#define TSC2000_REG_Z2		((0 << 11) | (3 << 5))
#define TSC2000_REG_BAT1	((0 << 11) | (5 << 5))
#define TSC2000_REG_BAT2	((0 << 11) | (6 << 5))
#define TSC2000_REG_AUX1	((0 << 11) | (7 << 5))
#define TSC2000_REG_AUX2	((0 << 11) | (8 << 5))
#define TSC2000_REG_TEMP1	((0 << 11) | (9 << 5))
#define TSC2000_REG_TEMP2	((0 << 11) | (0xA << 5))
#define TSC2000_REG_DAC		((0 << 11) | (0xB << 5))
#define TSC2000_REG_ZERO	((0 << 11) | (0x10 << 5))
#define TSC2000_REG_ADC		((1 << 11) | (0 << 5))
#define TSC2000_REG_DACCTL	((1 << 11) | (2 << 5))
#define TSC2000_REG_REF		((1 << 11) | (3 << 5))
#define TSC2000_REG_RESET	((1 << 11) | (4 << 5))
#define TSC2000_REG_CONFIG	((1 << 11) | (5 << 5))

#define TSC2000_DELAY_BASE	500

#define TSC2000_NO_SENSOR	-0x10000

#define SLEEP_TIMEOUT		5	/* ticks */

#include <linux/interrupt.h>

static unsigned short tsc2000_read (unsigned short reg);
static void tsc2000_write (unsigned short reg, unsigned short data);
static void tsc2000_reg_init (void);

static int tsc2000_spi_init = 0;
static int tsc2000_chrdev_registered = 0;
static int tsc2000_thread_started = 0;
static devfs_handle_t tsc2000_devfs_dir = 0;
#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *tsc2000_proc_dir = 0;
#endif

static int tsc2000_thread_quit;
static atomic_t tsc2000_thread_is_event;
static struct semaphore tsc2000_thread_start;
static struct semaphore tsc2000_thread_exit;
static wait_queue_head_t tsc2000_thread_event;

static struct semaphore tsc2000_lock;
#if defined(CONFIG_TRAB_TOUCH)
extern void stop_touch(void *);
#endif
#ifndef NEWCPLD
#define cpld_touch  ((unsigned long *)io_p2v(0x04010000))
#define INTR_DAV	(1 << 16)
#define INTR_TOUCH	(1 << 17)
#endif
atomic_t tsc2000_eint0;
static DECLARE_WAIT_QUEUE_HEAD(tsc2000_dav_wq);

static struct offs_amp
{
	int range1_2;
	int range3;
} tsc2000_offset_error;

/* How many values store to compute average of
 * the amplifier offset error.
 */
#define STORE_NUM 10

/* The period in sec to measure the amplifier 
 * offset error.
 */
#define OFFSET_MEASURE_PERIOD  10
/* The period in sec to measure pt1000 bridge
 * contact voltage
 */
#define PT1000_MEASURE_PERIOD  2
/* precomputed error correction factors, rounded */
static long x_range[3];
/* offset of the pt1000 bridge, nV */
static int tsc2000_offset_pt1000;
/* battery error, in o/oo */
static int tsc2000_error_battery;

/* Aux variables to avoid frequent measurements of the 
 * Pt1000 voltage.
 */
static int Upt1000_measured;
static unsigned long Upt1000_jiffies;
static long Upt1000_latched;

/* initial values for error correction factors */
#define TSC2000_ERR_VREF   0  /* ppm */
#define TSC2000_ERR_AMP1   0  /* ppm */
#define TSC2000_ERR_AMP2   0  /* ppm */
#define TSC2000_ERR_AMP3   0  /* ppm */
#define TSC2000_ERR_OFFSET 0  /* uV  */
#define TSC2000_ERR_BATTERY 0  /* o/oo  */

/* global structure to keep current vaules of 
 * correction factors.
 */
static tsc2000_correction_t tsc2000_correction = 
{
	TSC2000_ERR_VREF,
	TSC2000_ERR_AMP1,
	TSC2000_ERR_AMP2,
	TSC2000_ERR_AMP3,
	TSC2000_ERR_OFFSET,
	TSC2000_ERR_BATTERY
};

typedef struct tsc2000_sensor_s {
	const char *name;
	int auto_check;
	s32 (*measure) (unsigned int);
	unsigned int index;
	unsigned int minor;

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *proc_entry;
#endif
	devfs_handle_t devfs_entry;
	s32 min;
	s32 max;
	u32 delta;
	s32 value;
	u32 interval;
	wait_queue_head_t waitq;
	unsigned int event;
	unsigned long time_left;
	unsigned long last_measured;
} tsc2000_sensor_t;

#ifdef CONFIG_PROC_FS
typedef struct tsc2000_proc_s {
	tsc2000_sensor_t *sensor;
	struct semaphore lock;
	unsigned int inpos;
	unsigned int insize;
	unsigned int outpos;
	char indata [256];
	char outdata [256];
} tsc2000_proc_t;
#endif

/* atomic_set() and atomic_read() are not atomic !!! */
#define my_atomic_read(x, v) { \
	unsigned long flags; \
	save_flags(flags); \
	cli(); \
	(x) = ((v)->counter); \
	restore_flags(flags); \
}

#define my_atomic_set(v, i) { \
	unsigned long flags; \
	save_flags(flags); \
	cli(); \
	((v)->counter) = (i); \
	restore_flags(flags); \
}

#ifndef CONFIG_TRAB_TOUCH
static inline
#endif
void spi_wait_transmit_done (void)
{
	unsigned int i = 0;

	for (i = 0; i < 100; i++) {
		if (rSPSTA & 0x01) {
			return;
		}
		udelay(1);
	}

	printk (KERN_ERR TSC2000_NAME
		": Timeout while waiting for SPI transmit done\n");
}

static int __init spi_init (void)
{
	unsigned int i;

	/* Configure I/O ports. */
	rPDCON = (rPDCON & 0xF3FFFF) | 0x040000;
	rPGCON = (rPGCON & 0x0F3FFF) | 0x008000;
	rPGCON = (rPGCON & 0x0CFFFF) | 0x020000;
	rPGCON = (rPGCON & 0x03FFFF) | 0x080000;
	
	CLR_CS_TOUCH;

#if 0
	/* Use GPG9,8,7 as SPI_CLK, SPI-MOSI, SPI-MISO */
	rPGCON = (rPGCON & 0x03FFF) | 0x0A8000;
#endif

	/* Baudrate ca. 514kHz. */
	rSPPRE = 0x1F;

	/* SPI-MOSI holds Level after last bit. */
	rSPPIN = 0x01;

	/* Polling mode, Prescaler, Master, CPOL=0, CPHA=1. */
	rSPCON = 0x1A;

	/* Dummy byte ensures clock to be low. */
	for (i = 0; i < 10; i++) {
		rSPTDAT = 0xFF;
	}

	spi_wait_transmit_done();

	return 0;
}

static void spi_exit (void)
{
	spi_wait_transmit_done();
}

static struct callout_list *tsc2000_callout_handle = NULL;

static void tsc2000_reset(void)
{
	tsc2000_write(TSC2000_REG_RESET, 0xBBFF);
	udelay(10);     /* just in case */

		/* Re-initilize the registers again.
		 */
	tsc2000_reg_init();
}

static void dav_timeout(void *junk)
{
	DEBUG_CLR_PDDAT_0;
	tsc2000_reset();
	my_atomic_set(&tsc2000_eint0, 0);
	wake_up_interruptible(&tsc2000_dav_wq);
	printk (KERN_WARNING TSC2000_NAME ": dav_timeout()\n");
}

static void tsc2000_dav_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
	atomic_t *ap = (atomic_t *)dev_id;
	int res;
#ifndef NEWCPLD
	int whichintr;

	whichintr = *cpld_touch;
	/* not IRQ_DAV */
	if ((whichintr & INTR_DAV) == 0)
		return;
#endif
	/* not expecting an interrupt */
	my_atomic_read(res, ap);
	if (res == 0)
		return;
#ifndef NEWCPLD
	/* clear the interrupt bits in the CPLD */
	*cpld_touch = (INTR_DAV|INTR_TOUCH);
#endif
	if (tsc2000_callout_handle != NULL)
	{
		rmv_callout(tsc2000_callout_handle);
		tsc2000_callout_handle = NULL;
	}
	wake_up_interruptible(&tsc2000_dav_wq);
}

static void tsc2000_write (unsigned short reg, unsigned short data)
{
	unsigned short command;

	SET_CS_TOUCH;
	command = reg;
	rSPTDAT = (command & 0xFF00) >> 8;
	spi_wait_transmit_done();
	rSPTDAT = (command & 0x00FF);
	spi_wait_transmit_done();
	rSPTDAT = (data & 0xFF00) >> 8;
	spi_wait_transmit_done();
	rSPTDAT = (data & 0x00FF);
	spi_wait_transmit_done();
	CLR_CS_TOUCH;
}

static unsigned short tsc2000_read (unsigned short reg)
{
	unsigned short command, data;

	SET_CS_TOUCH;
	command = 0x8000 | reg;

	rSPTDAT = (command & 0xFF00) >> 8;
	spi_wait_transmit_done();
	rSPTDAT = (command & 0x00FF);
	spi_wait_transmit_done();

	rSPTDAT = 0xFF;
	spi_wait_transmit_done();
	data = rSPRDAT;
	rSPTDAT = 0xFF;
	spi_wait_transmit_done();

	CLR_CS_TOUCH;
	return (rSPRDAT & 0x0FF) | (data << 8);
}

static void tsc2000_set_mux (unsigned int channel)
{
	switch (channel) {
	case 0:	
		CLR_MUX0; CLR_MUX1;
		SET_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 1:
		SET_MUX0; CLR_MUX1;
		SET_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 2:
		CLR_MUX0; SET_MUX1;
		SET_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 3:
		SET_MUX0; SET_MUX1;
		SET_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 4:
		CLR_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; SET_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 5:
		SET_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; SET_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 6:
		CLR_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; SET_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 7:
		SET_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; SET_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 8:
		CLR_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		SET_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 9:
		SET_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		SET_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 10:
		CLR_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		SET_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 11:
		SET_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		SET_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	case 12:
		CLR_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; SET_MUX4_ENABLE;
		break;
	case 13:
		SET_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; SET_MUX4_ENABLE;
		break;
	case 14:
		CLR_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; SET_MUX4_ENABLE;
		break;
	case 15:
		SET_MUX0; SET_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; SET_MUX4_ENABLE;
		break;
	default:
		CLR_MUX0; CLR_MUX1;
		CLR_MUX1_ENABLE; CLR_MUX2_ENABLE;
		CLR_MUX3_ENABLE; CLR_MUX4_ENABLE;
		break;
	}
}

static void tsc2000_set_range (unsigned int range)
{
	switch (range) {
	case 1:
		CLR_SEL_TEMP_V_0; SET_SEL_TEMP_V_1;
		CLR_SEL_TEMP_V_2; CLR_SEL_TEMP_V_3;
		break;
	case 2:
		CLR_SEL_TEMP_V_0; CLR_SEL_TEMP_V_1;
		CLR_SEL_TEMP_V_2; SET_SEL_TEMP_V_3;
		break;
	case 3:
		SET_SEL_TEMP_V_0; CLR_SEL_TEMP_V_1;
		SET_SEL_TEMP_V_2; CLR_SEL_TEMP_V_3;
		break;
	} 
}

static void tsc2000_reg_init (void)
{
	tsc2000_write(TSC2000_REG_ADC,    0x2036);
	tsc2000_write(TSC2000_REG_REF,    0x0000);
	tsc2000_write(TSC2000_REG_DACCTL, 0x0000);

	CON_MUX0;
	CON_MUX1;

	CON_MUX1_ENABLE;
	CON_MUX2_ENABLE;
	CON_MUX3_ENABLE;
	CON_MUX4_ENABLE;

	CON_SEL_TEMP_V_0;
	CON_SEL_TEMP_V_1;
	CON_SEL_TEMP_V_2;
	CON_SEL_TEMP_V_3;

	tsc2000_set_mux(0);
	tsc2000_set_range(0);
}

static inline void tsc2000_reg_exit (void)
{
	/* Nothing */
}

/* val must be in [0, 4095] */
static inline unsigned long tsc2000_to_uv (u16 val)
{
	return ((250000 * val) / 4096) * 10;
}

static u16 tsc2000_read_channel (unsigned int channel)
{
	unsigned long flags;
	u16 res;

	tsc2000_set_mux(channel);
	udelay(3 * TSC2000_DELAY_BASE);
	
	save_flags(flags);
	cli();

#ifdef CONFIG_TRAB_TOUCH
	stop_touch((void *)2);
#endif
	DEBUG_SET_PDDAT_0;
	tsc2000_write(TSC2000_REG_ADC, 0x2036);
	my_atomic_set(&tsc2000_eint0, 1);
	tsc2000_callout_handle = add_callout(dav_timeout, NULL, 2, 0);
	interruptible_sleep_on(&tsc2000_dav_wq);

	res = tsc2000_read(TSC2000_REG_AUX1);
	DEBUG_CLR_PDDAT_0;
	my_atomic_set(&tsc2000_eint0, 0);
	restore_flags(flags);
	return res;
}

static s32 tsc2000_measure_battery (unsigned int i)
{
	unsigned long flags;
	s32 ret, res;

	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}
	
	save_flags(flags);
	cli();

#ifdef CONFIG_TRAB_TOUCH
	stop_touch((void *)3);
#endif
	DEBUG_SET_PDDAT_4;
	tsc2000_write(TSC2000_REG_ADC, 0x1836);
	DEBUG_SET_PDDAT_0;
	my_atomic_set(&tsc2000_eint0, 1);
	tsc2000_callout_handle = add_callout(dav_timeout, NULL, 2, 0);
	interruptible_sleep_on(&tsc2000_dav_wq);
	ret = tsc2000_read(TSC2000_REG_BAT1);
	DEBUG_CLR_PDDAT_0;
	my_atomic_set(&tsc2000_eint0, 0);
	restore_flags(flags);

	up(&tsc2000_lock);
	res = (tsc2000_to_uv(ret) + 1250) / 2500;
	return res + (tsc2000_error_battery * res) / 1000;
}

static s32 tsc2000_measure_pressure (unsigned int i)
{
	s32 ret;
	unsigned long flags;

	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}

	save_flags(flags);
	cli();

#ifdef CONFIG_TRAB_TOUCH
	stop_touch((void *)4);
#endif
	DEBUG_SET_PDDAT_4;
	tsc2000_write(TSC2000_REG_ADC, 0x2436);
	DEBUG_SET_PDDAT_0;
	my_atomic_set(&tsc2000_eint0, 1);
	tsc2000_callout_handle = add_callout(dav_timeout, NULL, 2, 0);
	interruptible_sleep_on(&tsc2000_dav_wq);
	ret = tsc2000_read(TSC2000_REG_AUX2);
	DEBUG_CLR_PDDAT_0;
	my_atomic_set(&tsc2000_eint0, 0);
	restore_flags(flags);

	up(&tsc2000_lock);

/* !!! BUG: the value needs to be converted to TBD */
	return ret;
}

/* Update correction factors, recalculate constants */
static inline void tsc2000_update_corr(void)
{
	if (down_interruptible(&tsc2000_lock))
		return;

	/* error correction factors for each metering range, rounded */
	x_range[0] = 
		((unsigned long)(2500 * (1000000ul +
			tsc2000_correction.error_vref +
		tsc2000_correction.error_amp1)) + 4095 * 2) / (4095 * 4);
	x_range[1] = 
		((unsigned long)(2500 * (1000000ul +
			tsc2000_correction.error_vref +
		tsc2000_correction.error_amp2)) + 4095 * 2) / (4095 * 4);
	x_range[2] = 
		((unsigned long)(2500 * (1000000ul +
		tsc2000_correction.error_vref +
		tsc2000_correction.error_amp3)) + 4095 * 3) / (4095 * 6);

	/* Convert from user-supplied uV to nV */
	tsc2000_offset_pt1000 = tsc2000_correction.offset_pt1000 * 1000;

	tsc2000_error_battery = tsc2000_correction.error_battery;
#ifdef DEBUG
	printk(KERN_INFO "%s: Correction factors:x_range1 %ld, x_range2 %ld,"
		   " x_range3 %ld, pt1000 offset %d, battery voltage %d\n",
		   __FUNCTION__, x_range[0], x_range[1], x_range[2], 
		   tsc2000_offset_pt1000, tsc2000_error_battery);
#endif
	up(&tsc2000_lock);
}

static void tsc2000_moving_average(void)
{
	static int values1_2[STORE_NUM];
	static int values3[STORE_NUM];
	static int inited = 0;
	static int replace = 0;
	static int sum1_2;
	static int sum3;
	int i;
	
	if (!inited) {
		for (i = 0; i < STORE_NUM; i++)
			values1_2[i] = tsc2000_offset_error.range1_2;
		for (i = 0; i < STORE_NUM; i++)
			values3[i] = tsc2000_offset_error.range3;
		sum1_2 = STORE_NUM * tsc2000_offset_error.range1_2;
		sum3 = STORE_NUM * tsc2000_offset_error.range3;
		inited++;
	} else {
		sum1_2 -= values1_2[replace];
		sum1_2 += tsc2000_offset_error.range1_2;
		sum3 -= values3[replace];
		sum3 += tsc2000_offset_error.range3;

		values1_2[replace] = tsc2000_offset_error.range1_2;
		values3[replace] = tsc2000_offset_error.range3;		
		replace = (replace + 1) % STORE_NUM;
		tsc2000_offset_error.range1_2 = sum1_2 / STORE_NUM;
		tsc2000_offset_error.range3 = sum3 / STORE_NUM;
	}
}

/* The function is called periodically (10s) to monitor
 * the differential amplifier offset error.
 */
static void tsc2000_update_offset_error(void)
{
	unsigned long scan;
	static int offset_error_measured = 0;
	static unsigned long offset_error_jiffies;

	if (!offset_error_measured || jiffies > offset_error_jiffies +
		OFFSET_MEASURE_PERIOD*HZ) {
		offset_error_measured = 1;
		offset_error_jiffies = jiffies;

		/* Is this sensor always present? */
		tsc2000_set_range(1);
		scan = tsc2000_read_channel(15);

		if (scan < 4095) {
			tsc2000_offset_error.range1_2 = 4095 - scan;
		} else {
			tsc2000_set_range(2);
			tsc2000_offset_error.range1_2 = tsc2000_read_channel(15);
		}
		tsc2000_set_range(3);
		tsc2000_offset_error.range3 = tsc2000_read_channel(15);
		tsc2000_moving_average();
	}

	return;
}

static s32 tsc2000_measure_voltage (unsigned int channel,
		unsigned int range, unsigned long *adc_raw)
{
	long voltage = 0;
	long scan;

	tsc2000_update_offset_error();
	tsc2000_set_range(range);
	scan = tsc2000_read_channel(channel);
	switch (range) {
	case 1:
		voltage = (x_range[range-1] *
			((scan - tsc2000_offset_error.range1_2) - 4095)) / 100;
		break;
	case 2:
		voltage = (x_range[range-1] *
			(scan - tsc2000_offset_error.range1_2)) / 100;
		break;
	case 3:
		voltage = (x_range[range-1] *
			(scan - tsc2000_offset_error.range3)) / 10;
		break;
	}

	/* Return raw ADC data to the caller, if required */
	if (adc_raw)
		*adc_raw = scan;
#ifdef DEBUG
	printk(KERN_INFO "%s: ch %d, range %d -> voltage %ld (adc_raw %ld,"
		   "offset error %d, x_range %ld)\n",
		   __FUNCTION__, channel, range, voltage, scan,
		   (range == 3) ? tsc2000_offset_error.range3 :
		   tsc2000_offset_error.range1_2, x_range[range-1]);
#endif
	return voltage;
}

static int tsc2000_interpolate(long value, long data[][2], long *result)
{
	int i;

	/* the data is sorted and the first element is upper
	 * limit so we can easily check for out-of-band values 
	 */
	if (data[0][0] < value || data[1][0] > value)
		return -1;

	i = 1;
	while (data[i][0] < value)
		i++;

	/* To prevent overflow we have to store the intermediate 
	   result in 'long long'.
	*/
	*result = data[i-1][1] +
		((unsigned long long)(data[i][1] - data[i-1][1])
		 * (unsigned long long)(value - data[i-1][0]))
		/ (data[i][0] - data[i-1][0]);

	return 0;
}

static s32 tsc2000_measure_thermoel (unsigned int channel)
{
	unsigned long adc_raw;
	long Ue, Upt1000, Ut, Utsaved, thermo_temp, Uc;
	static int ranges_used[14] = {[0 ... 13] = 3};

	int last_range = ranges_used[channel];

	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}

	if (Upt1000_measured && jiffies < Upt1000_jiffies +
		PT1000_MEASURE_PERIOD*HZ) {
		Upt1000 = Upt1000_latched;
	} else {
		Upt1000_measured = 1;
		Upt1000_jiffies = jiffies;
		Upt1000 = tsc2000_measure_voltage(14, 3, NULL);
		Upt1000_latched = Upt1000;
	}

	Upt1000 += tsc2000_offset_pt1000;
	if (tsc2000_interpolate(Upt1000,
							Pt1000_voltage_table, &Uc) == -1) {
		printk (KERN_ERR TSC2000_NAME
			": interpolation failed for %ld\n", Upt1000);
		up(&tsc2000_lock);
		return TSC2000_NO_SENSOR;
	}

	Ut = tsc2000_measure_voltage(channel, last_range, &adc_raw);

	switch (last_range) {
	case 1:
	case 2:
		if (adc_raw != 0 && adc_raw != 4095)
			break;
		else {
#ifdef DEBUG
			printk(KERN_INFO "%s: switching to range 3 (adc %ld\n)", 
				   __FUNCTION__, adc_raw);
#endif
			last_range = 3;
			Ut = tsc2000_measure_voltage(channel, last_range, &adc_raw);
		}
		/* Intentional fall through */
	case 3:
		if (adc_raw == 4095) {
			up(&tsc2000_lock);
			return TSC2000_NO_SENSOR;				
		}
		Utsaved = Ut;
		if (Ut >= -6200000 && Ut < 0) {
#ifdef DEBUG
			printk(KERN_INFO "%s: switching to range 1 (Ut %ld)\n", 
				   __FUNCTION__, Ut);
#endif
			Ut = tsc2000_measure_voltage(channel, 1, &adc_raw);
			if (adc_raw == 0) {
				Ut = Utsaved;
			} else {
				last_range = 1;
				if (adc_raw == 4095) {
					/* overflow, switch to range 2 */
#ifdef DEBUG
			printk(KERN_INFO "%s: overflow, switching to range 2\n", 
				   __FUNCTION__);
#endif
					Ut = tsc2000_measure_voltage(channel, 
						2, &adc_raw);
					last_range = 2;
				}
			}
		} else 
			if (Ut >= 0 && Ut < 6200000) {
				/* switch to range 2 */
#ifdef DEBUG
				printk(KERN_INFO "%s: switching to range 2"
					" (Ut %ld)\n", __FUNCTION__, Ut);
#endif
				Ut = tsc2000_measure_voltage(channel, 2,
					&adc_raw);
				if (adc_raw == 4095) {
					Ut = Utsaved;
				} else {
					last_range = 2;
					if (adc_raw == 0) {
#ifdef DEBUG
						printk(KERN_INFO "%s: neg overflow,"
							   "switching to range 1\n",
							   __FUNCTION__);
#endif
						/* switch to range 1 */
						Ut = tsc2000_measure_voltage(channel,
							1, &adc_raw);
						last_range = 1;
					}
				}
			}
		break;
	}
	up(&tsc2000_lock);

	ranges_used[channel] = last_range;
	/* Now add the contact voltage */
	Ue = Ut + Uc;
	if (tsc2000_interpolate(Ue, TypK_temp_table, &thermo_temp) == -1) {
		printk (KERN_ERR TSC2000_NAME
			": interpolation failed for %ld\n", Ue);
		return TSC2000_NO_SENSOR;
	}
#ifdef DEBUG
	printk(KERN_INFO "%s: Voltage: thermopair %ld,"
		   "contact %ld, resulted temp %ld\n",
		   __FUNCTION__, Ut, Uc, thermo_temp);
#endif
	return thermo_temp;
}

static s32 tsc2000_measure_contact_temp (unsigned int i)
{
	long contact_temp;
	long Upt1000;

	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}
	
	Upt1000 = tsc2000_measure_voltage(14, 3, NULL);
	Upt1000_measured = 1;
	Upt1000_latched = Upt1000;
	Upt1000_jiffies = jiffies;

	Upt1000 += tsc2000_offset_pt1000;
	if (tsc2000_interpolate(Upt1000, Pt1000_temp_table, 
					&contact_temp) == -1) {
		printk (KERN_ERR TSC2000_NAME
			": interpolation failed for %ld\n", Upt1000);
		contact_temp = TSC2000_NO_SENSOR;
	}
	up(&tsc2000_lock);

	return contact_temp;
}

static s32 tsc2000_measure_brightness (unsigned int i)
{
	s32 brightness;
	
	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}

	brightness = tsc2000_read(TSC2000_REG_DAC);

	up(&tsc2000_lock);

	return brightness & 0xff;
}

#ifdef ADC_DEBUG
static s32 tsc2000_debug (unsigned int i)
{
	s32 debug_status = 0;

	if (down_interruptible(&tsc2000_lock)) {
		return TSC2000_NO_SENSOR;
	}

#ifdef CONFIG_TRAB_TOUCH
	stop_touch((void *)5);
#endif
		/* Reset TSC2000.
		 */
	tsc2000_reset();

	up(&tsc2000_lock);

	return debug_status;
}
#endif

/* Unspecified fields are zero initialized */
static tsc2000_sensor_t tsc2000_sensor [] = {
	{
		name: "temp0",
		auto_check: 1,
		index: 0,
		minor: TSC2000_MINOR_TEMP0,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp1",
		auto_check: 1,
		index: 1,
		minor: TSC2000_MINOR_TEMP1,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp2",
		auto_check: 1,
		index: 2,
		minor: TSC2000_MINOR_TEMP2,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp3",
		auto_check: 1,
		index: 4,
		minor: TSC2000_MINOR_TEMP3,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp4",
		auto_check: 1,
		index: 5,
		minor: TSC2000_MINOR_TEMP4,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp5",
		auto_check: 1,
		index: 6,
		minor: TSC2000_MINOR_TEMP5,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp6",
		auto_check: 1,
		index: 8,
		minor: TSC2000_MINOR_TEMP6,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp7",
		auto_check: 1,
		index: 9,
		minor: TSC2000_MINOR_TEMP7,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp8",
		auto_check: 1,
		index: 10,
		minor: TSC2000_MINOR_TEMP8,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp9",
		auto_check: 1,
		index: 12,
		minor: TSC2000_MINOR_TEMP9,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "temp10",
		auto_check: 1,
		index: 13,
		minor: TSC2000_MINOR_TEMP10,
		measure: tsc2000_measure_thermoel
	},
	{
		name: "battery",
		auto_check: 0,
		index: 0,
		minor: TSC2000_MINOR_BATTERY,
		measure: tsc2000_measure_battery
	},
	{
		name: "pressure",
		auto_check: 1,
		index: 0,
		minor: TSC2000_MINOR_PRESSURE,
		measure: tsc2000_measure_pressure
	},
	{
		name: "contact_temp",
		auto_check: 1,
		index: 0,
		minor: TSC2000_MINOR_CONTACT_TEMP,
		measure: tsc2000_measure_contact_temp
	},
	{
		name: "brightness",
		auto_check: 0,
		index: 0,
		minor: TSC2000_MINOR_BRIGHTNESS,
		measure: tsc2000_measure_brightness
	},
#ifdef ADC_DEBUG
#define TSC2000_MINOR_DEBUG	64
	{
		name: "debug",
		auto_check: 0,
		index: 0,
		minor: TSC2000_MINOR_DEBUG,
		measure: tsc2000_debug
	},
#endif
};
static size_t tsc2000_sensors = sizeof(tsc2000_sensor) /
	sizeof(tsc2000_sensor_t);

static ssize_t tsc2000_devfs_read (struct file *file, char *buf,
	size_t length, loff_t *pos)
{
	tsc2000_sensor_t *sensor = (tsc2000_sensor_t *)file->private_data;
	tsc2000_data_t data;
	int err = 0;
	unsigned int len = 0;
	
	if (! err) {
		if (pos != &file->f_pos) {
			err = -ESPIPE;
		}
	}

	if (! err) {
		/* To conform the API */
		if (length < sizeof(data)) {
			err = -ENOMEM;
		}
	}

	if (! err) {
		if ((file->f_flags & O_NONBLOCK) && ! sensor->auto_check) {
			err = -EAGAIN;
		}
	}

	if (! err) {
		if (sensor->auto_check) {
			data.value = sensor->value;
		} else {
			data.value = sensor->measure(sensor->index);
		}
	}

	if (! err) {
		len = length < sizeof(data) ? length : sizeof(data);
		if (copy_to_user((void *)buf, &data, len)) {
			err = -EFAULT;
		}
	}
	
	return err ? err : len;
}

static int tsc2000_devfs_ioctl (struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	tsc2000_sensor_t *sensor = (tsc2000_sensor_t *)file->private_data;
	int err = 0;
	tsc2000_data_t data;

	if (! err) {
		switch (cmd) {
		case TSC2000_IOCTL_SET_MIN:
			if (! err) {
				if (! sensor->auto_check) {
					err = -EINVAL;
				}
			}

			/* No check for R/O access to conform the API */

			if (! err) {
				if (copy_from_user(&data, (void *)arg,
					sizeof(data))) {
					err = -EFAULT;
				} else {
					sensor->min = data.value;
				}
			}
			break;
		case TSC2000_IOCTL_SET_MAX:
			if (! err) {
				if (! sensor->auto_check) {
					err = -EINVAL;
				}
			}
			
			/* No check for R/O access to conform the API */

			if (! err) {
				if (copy_from_user(&data, (void *)arg,
					sizeof(data))) {
					err = -EFAULT;
				} else {
					sensor->max = data.value;
				}
			}
			break;
		case TSC2000_IOCTL_SET_DELTA:
			if (! err) {
				if (! sensor->auto_check) {
					err = -EINVAL;
				}
			}

			/* No check for R/O access to conform the API */
			
			if (! err) {
				if (copy_from_user(&data, (void *)arg,
					sizeof(data))) {
					err = -EFAULT;
				} else {
					sensor->delta = data.value;
				}
			}
			break;
		case TSC2000_IOCTL_SET_INTERVAL:
			if (! err) {
				if (! sensor->auto_check) {
					err = -EINVAL;
				}
			}
			
			/* No check for R/O access to conform the API */

			if (! err) {
				if (copy_from_user(&data, (void *)arg,
					sizeof(data))) {
					err = -EFAULT;
				}
			}

			if (! err) {
				if (data.value < 10) {
					err = -EINVAL;
				}
			}

			if (! err) {
				sensor->interval = data.value;
				sensor->time_left = 0;
				my_atomic_set(&tsc2000_thread_is_event, 1);
				wake_up(&tsc2000_thread_event);
			}
			break;
		case TSC2000_IOCTL_SET_BRIGHTNESS:			
			if (MINOR(inode->i_rdev) != TSC2000_MINOR_BRIGHTNESS) {
				err = -EINVAL;
			}
			/* No check for R/O access to conform the API */

			if (! err) {
				if (copy_from_user(&data, (void *)arg,
					sizeof(data))) {
					err = -EFAULT;
				}
			}

			if (! err) {
				if (data.value < 0) {
					err = -EINVAL;
				}
			}

			if (! err) {
				down(&tsc2000_lock);
				tsc2000_write(TSC2000_REG_DAC, data.value & 0xff);
				up(&tsc2000_lock);
			}
			break;
		case TSC2000_IOCTL_GET_VALUE:
			if (file->f_flags & O_NONBLOCK) {
				err = -EAGAIN;
			}
			
			if (! err) {
				data.value = sensor->measure(sensor->index);
			}

			if (! err && copy_to_user((void *)arg, &data,
				sizeof(data))) {
				err = -EFAULT;
			}
			break;
		case TSC2000_IOCTL_GET_STATUS:
			if (! err) {
				if (! sensor->auto_check) {
					err = -EINVAL;
				}
			}

			if (! err) {
				data.value = sensor->event;
				sensor->event = 0;
			}

			if (! err && copy_to_user((void *)arg, &data,
				sizeof(data))) {
				err = -EFAULT;
			}
			break;
		case TSC2000_IOCTL_SET_CORRECTION:
		case TSC2000_IOCTL_GET_CORRECTION:
			if (MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP0	&&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP1	&&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP2 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP3 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP4 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP5 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP6 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP7 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP8 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP9 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_TEMP10 &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_CONTACT_TEMP &&
				MINOR(inode->i_rdev) != TSC2000_MINOR_BATTERY) {
				err = -EINVAL;
			}
			/* No check for R/O access to conform the API */
			if (! err) {
				if (cmd == TSC2000_IOCTL_SET_CORRECTION) {
					if (copy_from_user(&tsc2000_correction,
						(void *)arg,
						sizeof(tsc2000_correction))) {
						err = -EFAULT;
					}
					if (! err) {
						tsc2000_update_corr();
					}
				} else { /* TSC2000_IOCTL_GET_CORRECTION */
					if (copy_to_user((void *)arg, 
						&tsc2000_correction,
						sizeof(tsc2000_correction))) {
						err = -EFAULT;
					}
				}
			}
			break;
		default:
			err = -EINVAL;
			break;
		}
	}

	return err;
}

static int tsc2000_devfs_open (struct inode *inode, struct file *file)
{
	tsc2000_sensor_t *sensor = 0;
	int err = 0;
	unsigned int i;

	if (! err) {
		for (i = 0; i < tsc2000_sensors; i++) {
			sensor = tsc2000_sensor + i;
			
			if (sensor->minor == MINOR(inode->i_rdev)) {
				break;
			}
		}
		if (i == tsc2000_sensors) {
			err = -ENXIO;
		}
	}
	
	if (! err) {
		file->private_data = (void *)sensor;
		MOD_INC_USE_COUNT;
	}
	
	return err;
}

static int tsc2000_devfs_release (struct inode *inode, struct file *file)
{
	tsc2000_sensor_t *sensor = (tsc2000_sensor_t *)
		file->private_data;

	wake_up_interruptible(&sensor->waitq);

	MOD_DEC_USE_COUNT;

	return 0;
}

static unsigned int tsc2000_devfs_poll(struct file *file, poll_table *wait)
{
	tsc2000_sensor_t *sensor = (tsc2000_sensor_t *)file->private_data;
	unsigned int mask = sensor->auto_check ? POLLIN : 0;

	poll_wait(file, &sensor->waitq, wait);

	if (sensor->event) {
		mask |= POLLPRI;
	}

	return mask;
}

static struct file_operations tsc2000_devfs_fops = {
	owner:		THIS_MODULE,
	read:		tsc2000_devfs_read,
	ioctl:		tsc2000_devfs_ioctl,
	open:		tsc2000_devfs_open,
	release:	tsc2000_devfs_release,
	poll:		tsc2000_devfs_poll,
};

#ifdef CONFIG_PROC_FS
static ssize_t tsc2000_proc_read (struct file *file, char *buf,
	size_t length, loff_t *pos)
{
	tsc2000_proc_t *tp = (tsc2000_proc_t *)file->private_data;
	int err = 0;
	unsigned int len = 0;
	int locked = 0;

	if (! err) {
		if (pos != &file->f_pos) {
			err = -ESPIPE;
		}
	}

	if (! err) {
		if (down_interruptible(&tp->lock)) {
			err = -EAGAIN;
		} else {
			locked = 1;
		}
	}

	if (! err) {
		len = length < tp->insize - tp->inpos ? length : tp->insize -
			tp->inpos;
		if (copy_to_user((void *)buf, tp->indata + tp->inpos, len)) {
			err = -EFAULT;
		}
	}
	
	if (! err) {
		tp->inpos += len;
	}

	if (locked) {
		up(&tp->lock);
	}

	return err ? err : len;
}

static ssize_t tsc2000_proc_write (struct file *file, const char *buf,
	size_t length, loff_t *pos)
{
	tsc2000_proc_t *tp = (tsc2000_proc_t *)file->private_data;
	int err = 0;
	unsigned int len = 0;
	int locked = 0;
	
	if (! err) {
		if (pos != &file->f_pos) {
			err = -ESPIPE;
		}
	}

	if (! err) {
		if (down_interruptible(&tp->lock)) {
			err = -EAGAIN;
		} else {
			locked = 1;
		}
	}

	if (! err) {
		len = length < sizeof(tp->outdata) - 1 - tp->outpos ? length :
			sizeof(tp->outdata) - 1 - tp->outpos;
		if (copy_from_user(tp->outdata + tp->outpos, (void *)buf,
			len)) {
			err = -EFAULT;
		}
	}
	
	if (! err) {
		tp->outpos += len;
	}

	if (locked) {
		up(&tp->lock);
	}

	return err ? err : len;
}

static int tsc2000_proc_open (struct inode *inode, struct file *file)
{
	struct proc_dir_entry *de =
		(struct proc_dir_entry *)inode->u.generic_ip;
	tsc2000_sensor_t *sensor = (tsc2000_sensor_t *)de->data;
	tsc2000_proc_t *tp = 0;
	int err = 0;
	s32 value = 0;
	
	if (! err) {
		tp = (tsc2000_proc_t *)kmalloc(sizeof(tsc2000_proc_t),
			GFP_KERNEL);
		if (! tp) {
			err = -ENOMEM;
		}
	}
	
	if (! err) {
		if ((file->f_flags & O_NONBLOCK) && ! sensor->auto_check) {
			err = -EAGAIN;
		}
	}

	if (! err) {
		if (sensor->auto_check) {
			value = sensor->value;
		} else {
			value = sensor->measure(sensor->index);
		}
	}

	if (! err) {
		tp->sensor = sensor;
		tp->inpos = 0;
		tp->outpos = 0;
		init_MUTEX(&tp->lock);

		tp->insize = snprintf(tp->indata, sizeof(tp->indata), "%d\n",
			value);
	}
	
	if (! err) {
		file->private_data = (void *)tp;
		MOD_INC_USE_COUNT;
	} else {
		if (tp) {
			kfree(tp);
		}
	}
	
	return err;
}

static int tsc2000_parse (s32 *value, const char *name, const char *data)
{
	s32 val = 0;
	unsigned int pos = 0;
	int ret = 1;
	
	if (ret) {
		unsigned int len = strlen(name);

		ret = strncmp(data + pos, name, len) == 0;
		if (ret) {
			pos += len;
		}
	}
	
	if (ret) {
		ret = strncmp(data + pos, "=", sizeof("=") - 1) == 0;
		if (ret) {
			pos += sizeof("=") - 1;
		}
	}

	if (ret) {
		char * end;
		
		val = simple_strtol(data + pos, &end, 0);
		if (end != data + pos) {
			pos = end - data;
		} else {
			ret = 0;
		}
	}
	
	if (ret) {
		while (isspace(data[pos])) {
			pos++;
		}
		ret = ! data[pos];
	}

	if (ret) {
		*value = val;
	}
	
	return ret;
}

static int tsc2000_parseu (u32 *value, const char *name, const char *data)
{
	u32 val = 0;
	unsigned int pos = 0;
	int ret = 1;
	
	if (ret) {
		unsigned int len = strlen(name);

		ret = strncmp(data + pos, name, len) == 0;
		if (ret) {
			pos += len;
		}
	}
	
	if (ret) {
		ret = strncmp(data + pos, "=", sizeof("=") - 1) == 0;
		if (ret) {
			pos += sizeof("=") - 1;
		}
	}

	if (ret) {
		char * end;
		
		val = simple_strtoul(data + pos, &end, 0);
		if (end != data + pos) {
			pos = end - data;
		} else {
			ret = 0;
		}
	}
	
	if (ret) {
		while (isspace(data[pos])) {
			pos++;
		}
		ret = ! data[pos];
	}

	if (ret) {
		*value = val;
	}
	
	return ret;
}

static int tsc2000_proc_release (struct inode *inode, struct file *file)
{
	tsc2000_proc_t *tp = (tsc2000_proc_t *)file->private_data;
	tsc2000_sensor_t *sensor = tp->sensor;
	
	if (tp->outpos != 0) {
		s32 value;
		u32 valueu;

		tp->outdata[tp->outpos] = 0;

		if (sensor->auto_check) {
			if (tsc2000_parse(&value, "min", tp->outdata)) {
				sensor->min = value;
			} else if (tsc2000_parse(&value, "max", tp->outdata)) {
				sensor->max = value;
			} else if (tsc2000_parseu(&valueu, "delta",
				tp->outdata)) {
				sensor->delta = valueu;
			} else if (tsc2000_parseu(&valueu, "interval",
				tp->outdata)) {
				if (valueu >= 10) {
					sensor->interval = valueu;
					sensor->time_left = 0;
					my_atomic_set(&tsc2000_thread_is_event, 1);
					wake_up(&tsc2000_thread_event);
				}
			}
		} else if (sensor->minor == TSC2000_MINOR_BRIGHTNESS &&
				tsc2000_parseu(&valueu, "brightness",
				tp->outdata)) {
					down(&tsc2000_lock);
					tsc2000_write(TSC2000_REG_DAC,
						valueu & 0xff);
					up(&tsc2000_lock);
		}
	}

	kfree(tp);
	file->private_data = 0;
	
	MOD_DEC_USE_COUNT;
	
	return 0;
}

static struct file_operations tsc2000_proc_fops = {
	owner:		THIS_MODULE,
	read:		tsc2000_proc_read,
	write:		tsc2000_proc_write,
	open:		tsc2000_proc_open,
	release:	tsc2000_proc_release,
};
#endif /* CONFIG_PROC_FS */

static void tsc2000_update (tsc2000_sensor_t *sensor, s32 value)
{
	if (value >= sensor->max) {
		sensor->event |= TSC2000_STATUS_MAX;
	}
	if (value <= sensor->min) {
		sensor->event |= TSC2000_STATUS_MIN;
	}
	if ((value <= sensor->value && (u32)(sensor->value - value) >=
		sensor->delta) || (value > sensor->value &&
		(u32)(value - sensor->value) >= sensor->delta)) {
		sensor->event |= TSC2000_STATUS_DELTA;
	}

	sensor->value = value;

	if (sensor->event) {
		wake_up(&sensor->waitq);
	}
}

static int tsc2000_monitor_thread (void *data)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long old_jiffies = jiffies;
	int res;

	daemonize();

	spin_lock_irq(&current->sigmask_lock);
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	sprintf(current->comm, TSC2000_NAME "d") ;
	
	add_wait_queue(&tsc2000_thread_event, &wait);

	up(&tsc2000_thread_start);

	while (1) {
		unsigned long timeout = MAX_SCHEDULE_TIMEOUT;
		unsigned long passed;
		unsigned int i;
		unsigned long new_jiffies;

		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			sensor->last_measured = -1;
			
			if (sensor->auto_check && sensor->interval != 0 &&
				sensor->time_left < timeout) {
				timeout = sensor->time_left;
			}
		}
		
		set_current_state(TASK_UNINTERRUPTIBLE);
		my_atomic_read(res, &tsc2000_thread_is_event);
		schedule_timeout(timeout);
		my_atomic_set(&tsc2000_thread_is_event, 0);

		if (tsc2000_thread_quit) {
			break;
		}

		new_jiffies = jiffies;
		passed = new_jiffies - old_jiffies;
		old_jiffies = new_jiffies;
#ifdef DEBUG
		printk( KERN_INFO "jiffies: %ld; passed: %ld\n",
			jiffies, passed);
#endif
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			if (sensor->auto_check && sensor->interval != 0) {
				if (sensor->time_left == 0) {
					sensor->value = sensor->measure
						(sensor->index);
					sensor->time_left = sensor->interval;
					sensor->last_measured = jiffies;
				} else if (sensor->time_left <= passed) {
#ifdef DEBUG
				        printk( KERN_INFO
						"measure sensor: %s; "
						"jiffies: %ld; "
						"time_left: %ld; "
						"passed: %ld",
						sensor->name, jiffies,
						sensor->time_left, passed);
#endif
					sensor->time_left = sensor->interval;
#ifdef DEBUG
					printk( KERN_INFO
						"time_left: %ld\n",
						sensor->time_left);
#endif
					tsc2000_update(sensor, sensor->measure
						(sensor->index));
					sensor->last_measured = jiffies;
				} else {
					sensor->time_left -= passed;
				}
			}
		}
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			unsigned int delta = sensor->last_measured == -1 ? 0 :
				jiffies - sensor->last_measured;
			
			if (delta < sensor->time_left)
			{
				sensor->time_left -= delta;
			}
			else
			{
				sensor->time_left = 1;
			}
		}
			
	}

	remove_wait_queue(&tsc2000_thread_event, &wait);

	up(&tsc2000_thread_exit);
	
	return 0;
}

static void tsc2000_cleanup (void)
{
	unsigned int i;

	if (tsc2000_devfs_dir) {
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			if (sensor->devfs_entry) {
				devfs_unregister(sensor->devfs_entry);
				sensor->devfs_entry = 0;
			}
		}
	}

	if (tsc2000_devfs_dir) {
		devfs_unregister(tsc2000_devfs_dir);
		tsc2000_devfs_dir = 0;
	}

#ifdef CONFIG_PROC_FS
	if (tsc2000_proc_dir) {
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			if (sensor->proc_entry) {
				remove_proc_entry(sensor->name,
					tsc2000_proc_dir);
				sensor->proc_entry = 0;
			}
		}
	}

	if (tsc2000_proc_dir) {
		remove_proc_entry(TSC2000_NAME, NULL);
		tsc2000_proc_dir = 0;
	}
#endif

	if (tsc2000_chrdev_registered) {
		devfs_unregister_chrdev(TSC2000_MAJOR, TSC2000_NAME);
		tsc2000_chrdev_registered = 0;
#ifdef NEWCPLD
		free_irq(IRQ_EINT0, &tsc2000_eint0);
#else
		free_irq(IRQ_EINT3, &tsc2000_eint0);
#endif
	}

	if (tsc2000_thread_started) {
		my_atomic_set(&tsc2000_thread_is_event, 1);
		tsc2000_thread_quit = 1;
		wake_up(&tsc2000_thread_event);
		down(&tsc2000_thread_exit);
		tsc2000_thread_started = 0;
	}

	if (tsc2000_spi_init) {
		tsc2000_reg_exit();
		spi_exit();

		tsc2000_spi_init = 0;
	}
}

static int __init tsc2000_init (void)
{
	int err = 0;
	unsigned int i;

	if (! err) {
		spi_init();
		tsc2000_reg_init();
		init_MUTEX(&tsc2000_lock);

		tsc2000_spi_init = 1;
	}

	if (!err)
	{
#ifdef NEWCPLD /* use EINT0 for IRQ_DAV */
		if (request_irq(IRQ_EINT0, tsc2000_dav_interrupt, SA_SHIRQ, "EINT0", &tsc2000_eint0)) {
			printk("Failed to install an interrupt handler "
			       "for DAV IRQ%d\n", IRQ_EINT0);
			err = -EIO;
		}
		/* configure EINT0 as falling edge triggered */
		rEXTINT &= ~(7 << 0);
		rEXTINT |= (2 << 0);

		/* configure port pin as EINT0 */
		rPECON &= ~(3 << 0);
		rPECON |= (2 << 0);
#else /* use EINT3 for IRQ_DAV */
		if (request_irq(IRQ_EINT3, tsc2000_dav_interrupt, SA_SHIRQ, "EINT3", &tsc2000_eint0)) {
			printk("Failed to install an interrupt handler "
			       "for DAV IRQ%d\n", IRQ_EINT3);
			err = -EIO;
		}
		/* configure EINT3 as rising edge triggered */
		rEXTINT &= ~(7 << 12);
		rEXTINT |= (4 << 12);

		/* configure port pin as EINT3 */
		rPECON &= ~(3 << 6);
		rPECON |= (2 << 6);
		*cpld_touch = (INTR_DAV|INTR_TOUCH);
#endif
	}

	if (! err) {
		tsc2000_update_corr();
	}

	if (! err) {
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			init_waitqueue_head(&sensor->waitq);
			if (sensor->auto_check) {
				sensor->min = -0x80000000;
				sensor->max = 0x7fffffff;
				sensor->delta = 0xffffffff;
				sensor->interval = 0xffffffff;
				sensor->value = sensor->measure(sensor->index);
				sensor->event = 0;
				sensor->time_left = 0;
			}
		}
	}			

	if (! err) {
		init_waitqueue_head(&tsc2000_thread_event);
		init_MUTEX_LOCKED(&tsc2000_thread_exit);
		init_MUTEX_LOCKED(&tsc2000_thread_start);
		tsc2000_thread_quit = 0;
		my_atomic_set(&tsc2000_thread_is_event, 0);
		DEBUG_CLR_PDDAT_0;
		my_atomic_set(&tsc2000_eint0, 0);

		if (kernel_thread(tsc2000_monitor_thread, 0, 0) < 0) {
			printk (KERN_ERR TSC2000_NAME
				": Unable to start a thread !\n");
			err = -ENOMEM;
		} else {
			down(&tsc2000_thread_start);
			tsc2000_thread_started = 1;
		}
	}

	if (! err) {
		err = devfs_register_chrdev(TSC2000_MAJOR, TSC2000_NAME,
			&tsc2000_devfs_fops);
		
		tsc2000_chrdev_registered = ! err;
		
		if (! tsc2000_chrdev_registered) {
			printk (KERN_ERR TSC2000_NAME
				": Unable to register device !\n");
			err = -EIO;
		}
	}

#ifdef CONFIG_PROC_FS
	if (! err) {
		tsc2000_proc_dir = proc_mkdir(TSC2000_NAME, NULL);
		if (! tsc2000_proc_dir) {
			printk (KERN_ERR TSC2000_NAME
				": Unable to create /proc/%s directory !\n",
				TSC2000_NAME);
		}
	}

	if (! err && tsc2000_proc_dir) {
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			sensor->proc_entry = create_proc_entry(sensor->name,
				S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP |
				S_IWGRP | S_IROTH, tsc2000_proc_dir);

			if (sensor->proc_entry) {
				sensor->proc_entry->proc_fops =
					&tsc2000_proc_fops;
				sensor->proc_entry->data = (void *)sensor;
			} else {
				printk(KERN_ERR TSC2000_NAME
					": Unable to create "
					"/proc/%s/%s file !\n",
					TSC2000_NAME, sensor->name);
			}
		}
	}
#endif

	if (! err) {
		tsc2000_devfs_dir = devfs_mk_dir(NULL, TSC2000_NAME, NULL);
	}
	
	if (! err && tsc2000_devfs_dir) {
		for (i = 0; i < tsc2000_sensors; i++) {
			tsc2000_sensor_t *sensor = tsc2000_sensor + i;
			
			sensor->devfs_entry =
				devfs_register (tsc2000_devfs_dir,
				sensor->name, DEVFS_FL_DEFAULT, TSC2000_MAJOR,
				sensor->minor, S_IFREG | S_IRUSR | S_IWUSR |
				S_IRGRP | S_IWGRP | S_IROTH,
				&tsc2000_devfs_fops, NULL);
			if (! sensor->devfs_entry) {
				printk(KERN_ERR TSC2000_NAME
					": Unable to register the "
					"/dev/%s/%s file !\n",
					TSC2000_NAME, sensor->name);
			}
		}
	}
	
	if (err) {
		tsc2000_cleanup();
	} else {
		printk (KERN_INFO "TRAB TSC2000 A/D converter monitor "
			DRIVER_VERSION " started\n");
	}

	return err;
}

static void __exit tsc2000_exit (void) 
{
	tsc2000_cleanup();
}

module_init(tsc2000_init);
module_exit(tsc2000_exit);

MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("TSC2000");

EXPORT_NO_SYMBOLS;
