/*
 * I2C driver algorithms for S3C2400X
 * Copyright (c) 2002 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include <asm/arch/irqs.h>

#include <linux/i2c.h>
#include <linux/i2c-algo-s3c2400x.h>

#define rIICCON		(*(volatile unsigned long*)0xF5400000)
#define rIICSTAT	(*(volatile unsigned long*)0xF5400004)
#define rIICADD		(*(volatile unsigned long*)0xF5400008)
#define rIICDS		(*(volatile unsigned long*)0xF540000C)

#define rPFCON		(*(volatile unsigned long*)0xF5600038)
#define rPGCON		(*(volatile unsigned long*)0xF5600044)
#define rOPENCR		(*(volatile unsigned long*)0xF5600050)

static int i2c_scan = 0;
static int i2c_debug = 0;

static wait_queue_head_t i2c_wait;

static int arm_wait_for_completion (struct i2c_algo_s3c2400x_data *i2c_adap)
{
	int res = -1;

	if (!i2c_adap->poll) {
		int timeout;
		wait_queue_t __wait;

		init_waitqueue_entry (&__wait, current);

		add_wait_queue (&i2c_wait, &__wait);
		timeout = i2c_adap->int_sleep;
		while (1) {
			set_current_state (TASK_INTERRUPTIBLE);
			if (rIICCON & 0x10) {
				res = 0;
				break;
			}
			if (!timeout)
				break;
			timeout = schedule_timeout (timeout);
		}
		set_current_state (TASK_RUNNING);
		remove_wait_queue (&i2c_wait, &__wait);
	} else {
		int try;

		for (try = 0; try < i2c_adap->retry; try++) {
			if (rIICCON & 0x10) {
				res = 0;
				break;
			}
			udelay (i2c_adap->udelay);
		}
	}

	return res;
}

static int arm_wait_for_ack (struct i2c_algo_s3c2400x_data *i2c_adap)
{
	int res = -1;
	
	if (!arm_wait_for_completion (i2c_adap) && !(rIICSTAT & 0x01))
		res = 0;

	return res;
}

static int arm_wait_for_not_busy (struct i2c_algo_s3c2400x_data *i2c_adap)
{
	int try;
	int try_stop;
	int res = -1;

	for (try_stop = 0; try_stop < 1000; try_stop++) {
		for (try = 0; try < i2c_adap->retry; try++) {
			if ((rIICSTAT & 0x20) == 0) {
				res = 0;
				goto Done;
			}
			udelay (i2c_adap->udelay);
		}

		/*
		 * generate STOP signal
		 */
		rIICSTAT &= ~0x20;

		/*
		 * clear pending bit
		 */
		rIICCON &= ~0x10;
	}

  Done:

	return res;
}

static void arm_i2c_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
	if (i2c_debug > 1)
		printk ("\narm_i2c_interrupt(%d)\n", irq);

	/* Get 'me going again.
	 */
	wake_up_interruptible (&i2c_wait);
}

static void arm_i2c_init (struct i2c_algo_s3c2400x_data *i2c_adap)
{
	if (i2c_debug)
		printk ("%s()\n", __FUNCTION__);

	init_waitqueue_head (&i2c_wait);

	if (!i2c_adap->poll &&
		request_irq (IRQ_IIC, arm_i2c_interrupt, 0, "I2C",
					 (void *) i2c_adap)) {
		printk ("Failed to install an inetrrupt handler "
				"for I2C contorller IRQ%d\n", IRQ_IIC);

		i2c_adap->poll = 1;
	}

#ifndef	CONFIG_TRAB	/* SMDK2400 board */
	/*
	 * Configure PF1 as IICSDA, PF3 as IICSCL
	 */
	rPFCON |= 0xCC;

	/*
	 * Enable PF[1], PF[3] ports open-drain mode
	 */
	rOPENCR |= 0x03;
#else			/* TRAB board */
	/*
	 * Configure PG5 as IICSDA, PG6 as IICSCL
	 */
	rPGCON |= ((3 << 10) | (3 << 12));

	/*
	 * Enable PG[5], PG[6] ports open-drain mode
	 */
	rOPENCR |= 0x0C;
#endif	/* CONFIG_TRAB */

	/*
	 * set IIC-Bus transmit clock frequency
	 * enable ACK generation
	 * enable Tx/Rx interrupt
	 */
	rIICCON = 0xE0;	/* 64 kHz */
}


static int arm_i2c_shutdown (struct i2c_algo_s3c2400x_data *i2c_adap)
{
	if (i2c_debug)
		printk ("%s()\n", __FUNCTION__);

	arm_wait_for_not_busy (i2c_adap);

	rIICSTAT = 0;
	rIICCON = 0;

	if (!i2c_adap->poll) {
		free_irq (IRQ_IIC, (void *) i2c_adap);
	}

	return 0;
}

/* Write to IIC...
 * addr = address byte, with r/w flag already set
 */
static int
sendbytes (struct i2c_algo_s3c2400x_data *i2c_adap, u_char addr, char *buf,
		   int count)
{
	int b = 0;

	if (arm_wait_for_not_busy (i2c_adap) < 0) {
		return -1;
	}

	if (count == 0)
		return 0;

	/*
	 * enable ACK generation 
	 */
	rIICCON |= 0x80;

	/* set master receive mode
	 * enable serial output
	 */
	rIICSTAT = 0xD0;

	/*
	 * write slave address
	 */
	rIICDS = addr;

	/*
	 * generate START signal
	 */
	rIICSTAT = 0xF0;

	/*
	 * Wait for acknowledge
	 */
	if (arm_wait_for_ack (i2c_adap) < 0) {
		goto Done;
	}

	for (b = 0; b < count; b++) {
		/*
		 * write data
		 */
		rIICDS = buf[b];

		/*
		 * clear pending bit to resume
		 */
		rIICCON &= ~0x10;

		if (arm_wait_for_ack (i2c_adap) < 0) {
			goto Done;
		}
	}

  Done:

	/*
	 * generate STOP signal
	 */
	rIICSTAT = 0xD0;

	/*
	 * clear pending bit
	 */
	rIICCON &= ~0x10;

	udelay (10000);

	return b;
}

/* Read from IIC...
 * addr = address byte, with r/w flag already set
 */
static int
readbytes (struct i2c_algo_s3c2400x_data *i2c_adap, u_char addr, char *buf,
		   int count)
{
	int b = 0;

	if (count == 0)
		return 0;

	if (arm_wait_for_not_busy (i2c_adap) < 0) {
		return -1;
	}

	/*
	 * enable ACK generation 
	 */
	rIICCON |= 0x80;
	
	/* set master receive mode
	 * enable serial output
	 */
	rIICSTAT = 0x90;

	/*
	 * write slave address
	 */
	rIICDS = addr;

	/*
	 * generate START signal
	 */
	rIICSTAT = 0xB0;

	/*
	 * Wait for acknowledge
	 */
	if (arm_wait_for_ack (i2c_adap) < 0) {
		goto Done;
	}

	for (b = 0; b < count; b++) {
		/*
		 * generate NACK for the last byte
		 */
		if (b + 1 == count) {
			rIICCON &= ~0x80;
		}
		
		/*
		 * clear pending bit to resume
		 */
		rIICCON &= ~0x10;

		if (b + 1 == count) {
			/* no ACK is expected */
			if (arm_wait_for_completion (i2c_adap) < 0) {
				goto Done;
			}
		} else {
			/* an ACK is expected */
			if (arm_wait_for_ack (i2c_adap) < 0) {
				goto Done;
			}
		}

		/*
		 * read data
		 */
		buf[b] = rIICDS;
	}

  Done:

	/*
	 * generate STOP signal
	 */
	rIICSTAT = 0x90;

	/*
	 * clear pending bit
	 */
	rIICCON &= ~0x10;

	return b;
}

static int arm_xfer (struct i2c_adapter *adap,
					 struct i2c_msg msgs[], int num)
{
	struct i2c_algo_s3c2400x_data *arm_adap = adap->algo_data;
	struct i2c_msg *pmsg;
	int i, ret;
	u_char addr;

	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];

		if (i2c_debug)
			printk ("i2c-algo-s3c2400x.o: "
					"#%d addr=0x%x flags=0x%x len=%d\n",
					i, pmsg->addr, pmsg->flags, pmsg->len);

		addr = pmsg->addr << 1;
		if (pmsg->flags & I2C_M_RD)
			addr |= 1;

		if (pmsg->flags & I2C_M_RD) {
			/* read bytes into buffer */
			ret = readbytes (arm_adap, addr, pmsg->buf, pmsg->len);
			if (i2c_debug)
				printk ("i2c-algo-s3c2400x.o: read %d bytes\n", ret);
			if (ret < pmsg->len) {
				return (ret < 0) ? ret : -EREMOTEIO;
			}
		} else {
			/* write bytes from buffer */
			ret = sendbytes (arm_adap, addr, pmsg->buf, pmsg->len);
			if (i2c_debug)
				printk ("i2c-algo-s3c2400x.o: wrote %d\n", ret);
			if (ret < pmsg->len) {
				return (ret < 0) ? ret : -EREMOTEIO;
			}
		}
	}

	return num;
}

static int algo_control (struct i2c_adapter *adapter,
						 unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32 arm_func (struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_EMUL;
}

/* -----exported algorithm data: -------------------------------------	*/

static struct i2c_algorithm arm_algo = {
	"S3C2400X algorithm",
	I2C_ALGO_S3C2400X,
	arm_xfer,
	NULL,
	NULL,						/* slave_xmit       */
	NULL,						/* slave_recv       */
	algo_control,				/* ioctl        */
	arm_func,					/* functionality    */
};

/* 
 * registering functions to load algorithms at runtime 
 */
int i2c_s3c2400x_add_bus (struct i2c_adapter *adap)
{
	int i;
	struct i2c_algo_s3c2400x_data *arm_adap = adap->algo_data;

	if (i2c_debug)
		printk ("i2c-algo-s3c2400x.o: hw routines for %s registered.\n",
				adap->name);

	/* register new adapter to i2c module... */

	adap->id |= arm_algo.id;
	adap->algo = &arm_algo;

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	i2c_add_adapter (adap);
	arm_i2c_init (arm_adap);

	/* scan bus */
	if (i2c_scan) {
		u_char pos = 0;

		printk (KERN_INFO " i2c-algo-s3c2400x.o: scanning bus %s...\n",
				adap->name);
		for (i = 0x00; i < 0xFF; i += 2) {
			if (readbytes (arm_adap, i | 1, &pos, 1) == 1) {
				printk ("(%02x)", i);
			} else
				printk (".");
		}
		printk ("\n");
	}
	return 0;
}


int i2c_s3c2400x_del_bus (struct i2c_adapter *adap)
{
	int res;
	struct i2c_algo_s3c2400x_data *arm_adap = adap->algo_data;

	arm_i2c_shutdown (arm_adap);

	if ((res = i2c_del_adapter (adap)) < 0)
		return res;

	printk ("i2c-algo-s3c2400x.o: adapter unregistered: %s\n", adap->name);

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

EXPORT_SYMBOL (i2c_s3c2400x_add_bus);
EXPORT_SYMBOL (i2c_s3c2400x_del_bus);

int __init i2c_algo_s3c2400x_init (void)
{
	printk ("i2c-algo-s3c2400x.o: i2c s3c2400x algorithm module\n");
	return 0;
}

void __exit i2c_algo_s3c2400x_exit (void)
{
}

#ifdef MODULE
MODULE_AUTHOR ("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION ("I2C-Bus S3C2400X algorithm");

MODULE_PARM (i2c_scan, "i");
MODULE_PARM (i2c_debug, "i");

MODULE_PARM_DESC (i2c_scan, "Scan for active chips on the bus");
MODULE_PARM_DESC (i2c_debug,
				  "debug level - 0 off; 1 normal; 2 more verbose");
#endif

module_init (i2c_algo_s3c2400x_init);
module_exit (i2c_algo_s3c2400x_exit);
