/*
 * Mapping for the Samsung SMDK2400 board
 * (C) 2002 Wolfgang Denk, DENX Software Engineering, wd@denx.de
 *
 * This code was taken from iq80310.c and adapted for SMDK2400.
 *
 * Author:	Nicolas Pitre
 * Copyright:	(C) 2001 MontaVista Software Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

static struct mtd_info *mymtd;

static __u8 smdk2400_read8(struct map_info *map, unsigned long ofs)
{
	return *(__u8 *)(map->map_priv_1 + ofs);
}

static __u16 smdk2400_read16(struct map_info *map, unsigned long ofs)
{
	return *(__u16 *)(map->map_priv_1 + ofs);
}

static __u32 smdk2400_read32(struct map_info *map, unsigned long ofs)
{
	return *(__u32 *)(map->map_priv_1 + ofs);
}

static void smdk2400_copy_from(struct map_info *map, void *to,
		unsigned long from, ssize_t len)
{
	/*
	 * We can't use memcpy here due to the S3C2400X bug
	 * with loading exactly 2 registers from non-cached
	 * memory using the LDM instruction (depending on
	 * its arguments, the memcpy routine may do exactly
	 * that).
	 */
	while (len-- > 0) {
		*(volatile __u8 *)to++ = *(__u8 *)(map->map_priv_1 + from++);
	}
}

static void smdk2400_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	*(__u8 *)(map->map_priv_1 + adr) = d;
}

static void smdk2400_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	*(__u16 *)(map->map_priv_1 + adr) = d;
}

static void smdk2400_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	*(__u32 *)(map->map_priv_1 + adr) = d;
}

static struct map_info smdk2400_map = {
	name:		"SMDK2400 flash bank #0",
	size:		FLASH0_SIZE,
	buswidth:	4,
	read8:		smdk2400_read8,
	read16:		smdk2400_read16,
	read32:		smdk2400_read32,
	copy_from:	smdk2400_copy_from,
	write8:		smdk2400_write8,
	write16:	smdk2400_write16,
	write32:	smdk2400_write32,
};

static struct mtd_partition smdk2400_partitions[] = {
	{
		name:		"Firmware",
		size:		0x00200000,
		offset:		0,
		mask_flags:	MTD_WRITEABLE  /* force read-only */
	},{
		name:		"Filesystem",
		size:		0x00e00000,
		offset:		0x00200000
	}
};

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *mymtd;

static int __init init_smdk2400(void)
{
	if (!(smdk2400_map.map_priv_1 = (unsigned long)ioremap(FLASH0_BASE,
			FLASH0_SIZE))) {
		printk("%s: Failed to map %dMb of flash at 0x%x\n", __FUNCTION__,
			FLASH0_SIZE >> 20, FLASH0_BASE);
		return -EIO;
	}
	mymtd = do_map_probe("cfi_probe", &smdk2400_map);
	if (!mymtd) {
		return -ENXIO;
	}

	mymtd->module = THIS_MODULE;

	add_mtd_partitions(mymtd, smdk2400_partitions,
			NB_OF(smdk2400_partitions));

	return 0;
}

static void __exit cleanup_smdk2400(void)
{
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (smdk2400_map.map_priv_1) {
		iounmap((void *)smdk2400_map.map_priv_1);
	}
}

module_init(init_smdk2400);
module_exit(cleanup_smdk2400);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nicolas Pitre <nico@cam.org>");
MODULE_DESCRIPTION("MTD map driver for Samsung SMDK2400 board");
