/*
 * Mapping for the TRAB board
 * (C) 2002 Wolfgang Denk, DENX Software Engineering, wd@denx.de
 *
 * This code was taken from iq80310.c and adapted for TRAB
 *
 * Author:	Nicolas Pitre
 * Copyright:	(C) 2001 MontaVista Software Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

static struct mtd_info *mymtd;

static __u8 trab_read8(struct map_info *map, unsigned long ofs)
{
	return *(__u8 *)(map->map_priv_1 + ofs);
}

static __u16 trab_read16(struct map_info *map, unsigned long ofs)
{
	return *(__u16 *)(map->map_priv_1 + ofs);
}

static __u32 trab_read32(struct map_info *map, unsigned long ofs)
{
	return *(__u32 *)(map->map_priv_1 + ofs);
}

static void trab_copy_from(struct map_info *map, void *to,
		unsigned long from, ssize_t len)
{
	/*
	 * We can't use memcpy here due to the S3C2400X bug
	 * with loading exactly 2 registers from non-cached
	 * memory using the LDM instruction (depending on
	 * its arguments, the memcpy routine may do exactly
	 * that).
	 */
	while (len-- > 0) {
		*(volatile __u8 *)to++ = *(__u8 *)(map->map_priv_1 + from++);
	}
}

static void trab_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	*(__u8 *)(map->map_priv_1 + adr) = d;
}

static void trab_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	*(__u16 *)(map->map_priv_1 + adr) = d;
}

static void trab_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	*(__u32 *)(map->map_priv_1 + adr) = d;
}

static struct map_info trab_map = {
	name:		"TRAB flash bank #0",
	size:		FLASH0_SIZE,
	buswidth:	4,
	read8:		trab_read8,
	read16:		trab_read16,
	read32:		trab_read32,
	copy_from:	trab_copy_from,
	write8:		trab_write8,
	write16:	trab_write16,
	write32:	trab_write32,
};

static struct mtd_partition trab_partitions[] = {
	{
		name:		"Firmware 1",
		size:		0x00004000,
		offset:		0x00000000,
		mask_flags:	MTD_WRITEABLE  /* force read-only */
	},
	{
		name:		"Environment 1",
		size:		0x00004000,
		offset:		0x00004000,
	},
	{
		name:		"Environment 2",
		size:		0x00004000,
		offset:		0x00008000,
	},
	{
		name:		"Firmware 2",
		size:		0x00014000,
		offset:		0x0000C000,
		mask_flags:	MTD_WRITEABLE  /* force read-only */
	},
	{
		name:		"Kernel Image",
		size:		0x000E0000,
		offset:		0x00020000
	},
	{
		name:		"Filesystem 1",
		size:		0x00300000,
		offset:		0x00100000
	},
	{
		name:		"Filesystem 2",
		size:		0x00400000,
		offset:		0x00400000
	}
};

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *mymtd;

static int __init init_trab(void)
{
	if (!(trab_map.map_priv_1 = (unsigned long)ioremap(FLASH0_BASE,
			FLASH0_SIZE))) {
		printk("%s: Failed to map %dMb of flash at 0x%x\n",
			__FUNCTION__,
			FLASH0_SIZE >> 20, FLASH0_BASE);
		return -EIO;
	}
	mymtd = do_map_probe("cfi_probe", &trab_map);
	if (!mymtd) {
		return -ENXIO;
	}

	mymtd->module = THIS_MODULE;

	add_mtd_partitions(mymtd, trab_partitions,
			NB_OF(trab_partitions));

	return 0;
}

static void __exit cleanup_trab(void)
{
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (trab_map.map_priv_1) {
		iounmap((void *)trab_map.map_priv_1);
	}
}

module_init(init_trab);
module_exit(cleanup_trab);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nicolas Pitre <nico@cam.org>");
MODULE_DESCRIPTION("MTD map driver for TRAB board");
